# Copyright (c) 2015 Fortinet, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

#    FortiOS API request format templates.

# About api request message naming regulations:
# Prefix         HTTP method
# ADD_XXX    -->    POST
# SET_XXX    -->    PUT
# DELETE_XXX -->    DELETE
# GET_XXX    -->    GET

# Login
LOGIN = """
{
    "path": "/logincheck",
    "method": "POST",
    "body": {
        "username": "{{ username }}",
        "secretkey": "{{ secretkey }}"
    }
}
"""

RELOGIN = """login?redir=%2fapi%2fv2"""

LOGOUT = """
{
    "path": "/logout",
    "method": "POST"
}
"""

# Create VLAN
ADD_VLAN_INTERFACE = """
{
    "path": "/api/v2/cmdb/system/interface/",
    "method": "POST",
    "body": {
        "name": "interface",
        "json": {
            {% if name is defined %}
                "name": "{{ name }}",
            {% else %}
                "name": "os_vid_{{ vlanid }}",
            {% endif %}
            {% if vlanid is defined %}
                "vlanid": "{{ vlanid }}",
            {% endif %}
            "interface": "{{ interface }}",
            "vdom": "{{ vdom }}",
            "type": "vlan",
            {% if ip is defined %}
                "ip": "{{ ip }}",
                "mode": "static",
                "allowaccess": "ping",
            {% endif %}
            "secondary-IP":"enable",
            {% if alias is defined %}
                "alias": "{{ alias  }}",
            {% endif %}
            "ipv6": {
                "ip6-extra-addr": []
            }
        }
    }
}
"""

SET_VLAN_INTERFACE = """
{
    "path": "/api/v2/cmdb/system/interface/{{ name }}",
    "method": "PUT",
    "body": {
        "json": {
            {% if ip is defined and ip != None %}
                "ip": "{{ ip }}",
                "mode": "static",
                "allowaccess": "ping",
            {% endif %}
            {% if secondaryips is defined %}
                {% if secondaryips %}
                    "secondary-IP": "enable",
                    "secondaryip": [
                    {% for secondaryip in secondaryips[:-1] %}
                        {
                            "ip": "{{ secondaryip }}",
                            "allowaccess": "ping"
                        },
                    {% endfor %}
                        {
                            "ip": "{{ secondaryips[-1] }}",
                            "allowaccess": "ping"
                        }
                    ],
                {% else %}
                    "secondary-IP": "disable",
                {% endif %}
            {% endif %}
            {% if vlanid is defined %}
                "vlanid": "{{ vlanid }}",
            {% endif %}
            {% if vdom is defined %}
                "vdom": "{{ vdom }}"
            {% else %}
                "vdom": "root"
            {% endif %}
        }
    }
}
"""

# Delete VLAN (vlan id)
DELETE_VLAN_INTERFACE = """
{
    {% if vdom is defined %}
        "path":"/api/v2/cmdb/system/interface/{{ name }}/?vdom={{ vdom }}",
    {% else %}
        "path":"/api/v2/cmdb/system/interface/{{ name }}",
    {% endif %}
    "method": "DELETE"
}
"""

# Get VLAN interface info
GET_VLAN_INTERFACE = """
{
    {% if name is defined %}
        {% if vdom is defined %}
        "path":"/api/v2/cmdb/system/interface/{{ name }}/?vdom={{ vdom }}",
        {% else %}
        "path":"/api/v2/cmdb/system/interface/{{ name }}/",
        {% endif %}
    {% else %}
        {% if vdom is defined %}
            "path":"/api/v2/cmdb/system/interface/?vdom={{ vdom }}",
        {% else %}
            "path":"/api/v2/cmdb/system/interface/",
        {% endif %}
    {% endif %}
    "method": "GET"
}
"""


ADD_DHCP_SERVER = """
{
    "path":"/api/v2/cmdb/system.dhcp/server/",
    "method": "POST",
    "body": {
        "name": "server",
        {% if vdom is defined %}
        "vdom": "{{ vdom }}",
        {% endif %}
        "json": {
            "status":"enable",
            {% if dns_nameservers is defined and dns_nameservers %}
            "dns-service":"specify",
            {% for dns in dns_nameservers[:3] %}
            "dns-server{{ loop.index }}":"{{ dns }}",
            {% endfor %}
            {% else %}
            "dns-service":"default",
            {% endif %}
            {% if gateway != None %}
            "default-gateway":"{{ gateway }}",
            {% endif %}
            "netmask":"{{ netmask }}",
            "interface":"{{ interface }}",
            "ip-range":[
                {
                    "start-ip":"{{ start_ip }}",
                    "end-ip":"{{ end_ip }}"
                }
            ]
        }
    }
}
"""

SET_DHCP_SERVER = """
{
    {% if vdom is defined %}
        "path":"/api/v2/cmdb/system.dhcp/server/{{ id }}/?vdom={{ vdom }}",
    {% else %}
        "path":"/api/v2/cmdb/system.dhcp/server/{{ id }}",
    {% endif %}
    "method": "PUT",
    "body": {
        "json": {
            "status":"enable",
            {% if dns_nameservers is defined and dns_nameservers %}
            "dns-service":"specify",
            {% for dns in dns_nameservers[:3] %}
            "dns-server{{ loop.index }}":"{{ dns }}",
            {% endfor %}
            {% else %}
            "dns-service":"default",
            {% endif %}
            {% if gateway != None %}
            "default-gateway":"{{ gateway }}",
            {% endif %}
            "netmask":"{{ netmask }}",
            "interface":"{{ interface }}",
            "ip-range":[
                {
                    "start-ip":"{{ start_ip }}",
                    "end-ip":"{{ end_ip }}"
                }
            ]
        }
    }
}
"""


DELETE_DHCP_SERVER = """
{
    {% if vdom is defined %}
        "path":"/api/v2/cmdb/system.dhcp/server/{{ id }}/?vdom={{ vdom }}",
    {% else %}
        "path":"/api/v2/cmdb/system.dhcp/server/{{ id }}",
    {% endif %}
    "method": "DELETE"
}
"""

GET_DHCP_SERVER = """
{
    {% if id is defined %}
        {% if vdom is defined %}
            "path":"/api/v2/cmdb/system.dhcp/server/{{ id }}/?vdom={{ vdom }}",
        {% else %}
            "path":"/api/v2/cmdb/system.dhcp/server/{{ id }}",
        {% endif %}
    {% else %}
        {% if vdom is defined %}
            "path":"/api/v2/cmdb/system.dhcp/server/?vdom={{ vdom }}",
        {% else %}
            "path":"/api/v2/cmdb/system.dhcp/server",
        {% endif %}
    {% endif %}
    "method": "GET"
}
"""


SET_DHCP_SERVER_RSV_ADDR = """
{
    {% if vdom is defined %}
        "path":"/api/v2/cmdb/system.dhcp/server/{{ id }}/?vdom={{ vdom }}",
    {% else %}
        "path":"/api/v2/cmdb/system.dhcp/server/{{ id }}",
    {% endif %}
    "method": "PUT",
    "body": {
        "json": {
            "reserved-address": {{ reserved_address }}
        }
    }
}
"""


ADD_VDOM = """
{
    "path":"/api/v2/cmdb/system/vdom/",
    "method": "POST",
    "body": {
        "name": "vdom",
        "json": {
            "name":"{{ name }}"
        }
    }
}
"""

DELETE_VDOM = """
{
    "path":"/api/v2/cmdb/system/vdom/{{ name }}",
    "method": "DELETE"
}
"""

GET_VDOM = """
{
    "path":"/api/v2/cmdb/system/vdom/{{ name }}",
    "method": "GET"
}
"""

ADD_VDOM_LINK = """
{
    "path":"/api/v2/cmdb/system/vdom-link",
    "method": "POST",
    "body": {
        "json": {
            "name":"{{ name }}"
        }
    }
}
"""

DELETE_VDOM_LINK = """
{
    "path": "/api/v2/cmdb/system/vdom-link/{{ name }}",
    "method": "DELETE"
}
"""

GET_VDOM_LINK = """
{
    "path":"/api/v2/cmdb/system/vdom-link/{{ name }}",
    "method": "GET"
}
"""

ADD_ROUTER_STATIC = """
{
    "path": "/api/v2/cmdb/router/static",
    "method": "POST",
    "body": {
        {% if vdom is defined %}
            "vdom": "{{ vdom }}",
        {% else %}
            "vdom": "root",
        {% endif %}
        "json": {
            "dst": "{{ dst }}",
            "device": "{{ device }}",
            "gateway": "{{gateway }}"
        }
    }
}
"""

SET_ROUTER_STATIC = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/router/static/{{ id }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/router/static/{{ id }}",
    {% endif %}
    "method": "PUT",
    "body": {
        "json": {
            "dst": "{{ dst }}",
            "device": "{{ device }}",
            "gateway": "{{gateway }}"
        }
    }
}
"""


DELETE_ROUTER_STATIC = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/router/static/{{ id }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/router/static/{{ id }}",
    {% endif %}
    "method": "DELETE"
}
"""


GET_ROUTER_STATIC = """
{
    {% if id is defined %}
        {% if vdom is defined %}
            "path":"/api/v2/cmdb/router/static/{{ id }}/?vdom={{ vdom }}",
        {% else %}
            "path":"/api/v2/cmdb/router/static/{{ id }}/",
        {% endif %}
    {% else %}
        {% if vdom is defined %}
            "path":"/api/v2/cmdb/router/static/?vdom={{ vdom }}",
        {% else %}
            "path":"/api/v2/cmdb/router/static/",
        {% endif %}
    {% endif %}
    "method": "GET"
}
"""


ADD_FIREWALL_POLICY = """
{
    "path": "/api/v2/cmdb/firewall/policy/",
    "method": "POST",
    "body": {
        {% if vdom is defined %}
            "vdom": "{{ vdom }}",
        {% else %}
            "vdom": "root",
        {% endif %}
        "json": {
            "srcintf": [
                {
                    {% if srcintf is defined %}
                        "name": "{{ srcintf }}"
                    {% else %}
                        "name": "any"
                    {% endif %}
                }
            ],
            "dstintf": [
                {
                    {% if dstintf is defined %}
                        "name": "{{ dstintf }}"
                    {% else %}
                        "name": "any"
                    {% endif %}
                }
            ],
            "srcaddr":  [
                {
                    {% if srcaddr is defined %}
                        "name": "{{ srcaddr }}"
                    {% else %}
                        "name": "all"
                    {% endif %}
                }
            ],
            "dstaddr":  [
                {
                    {% if dstaddr is defined %}
                        "name": "{{ dstaddr }}"
                    {% else %}
                        "name": "all"
                    {% endif %}
                }
            ],
            {% if action is defined %}
                "action": "{{ action }}",
            {% else %}
                "action": "accept",
            {% endif %}
            "schedule": "always",
            {% if nat is defined %}
            "nat": "{{ nat }}",
            {% endif %}
            {% if poolname is defined %}
                {% if nat is not defined %}
                    "nat": "enable",
                {% endif %}
                "ippool": "enable",
                "poolname":[{
                    "name": "{{ poolname }}"
                }],
            {% endif %}
            {% if match_vip is defined %}
                "match-vip": "{{ match_vip }}",
            {% else %}
                "match-vip": "disable",
            {% endif %}
            {% if status is defined %}
                "status": "{{ status }}",
            {% else %}
                "status": "enable",
            {% endif %}
            "service":  [{
                {% if service is defined %}
                    "name": "{{ service }}"
                {% else %}
                    "name": "ALL"
                {% endif %}
            }],
            {% set profiles = {
                'av-profile': av_profile,
                'webfilter-profile': webfilter_profile,
                'ips-sensor': ips_sensor,
                'application-list': application_list,
                'ssl-ssh-profile': ssl_ssh_profile
            } %}
            {% set _utm_enable = true %}
            {% for k, v in profiles.iteritems() if v is defined and v %}
               {% if _utm_enable %}
                   {%set _utm_enable = false %}
                   "utm-status": "enable",
                   "profile-protocol-options":"default",
               {% endif %}
               "{{ k }}": "{{ v }}",
            {% else %}
               "utm-status": "disable",
               "profile-protocol-options": "",
            {% endfor %}
            {% if comments is defined %}
                "comments": "{{ comments }}"
            {% else %}
                "comments": ""
            {% endif %}
        }
    }
}
"""

SET_FIREWALL_POLICY = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall/policy/{{ id }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall/policy/{{ id }}",
    {% endif %}
    "method": "PUT",
    "body": {
        "json": {
            {% if srcintf is defined %}
                "srcintf": [
                    {
                        "name": "{{ srcintf }}"
                    }
                ],
            {% endif %}
            {% if dstintf is defined %}
                "dstintf": [
                    {
                        "name": "{{ dstintf }}"
                    }
                ],
            {% endif %}
            {% if srcaddr is defined %}
                "srcaddr":  [
                    {
                        "name": "{{ srcaddr }}"
                    }
                ],
            {% endif %}
            {% if dstaddr is defined %}
                "dstaddr":  [
                    {
                        "name": "{{ dstaddr }}"
                    }
                ],
            {% endif %}
            {% if action is defined %}
                "action": "{{ action }}",
            {% endif %}
            {% if nat is defined %}
            "nat": "{{ nat }}",
            {% endif %}
            {% if poolname is defined %}
                {% if nat is not defined %}
                    "nat": "enable",
                {% endif %}
                "ippool": "enable",
                "poolname":[{
                    "name":"{{ poolname }}"
                }],
            {% endif %}
            {% if match_vip is defined %}
                "match-vip":"{{ match_vip }}",
            {% endif %}
            {% if status is defined %}
                "status":"{{ status }}",
            {% endif %}
            {% if service is defined %}
                "service":  [{
                    "name": "{{ service }}"
                }],
            {% endif %}
            {% set profiles = {
                'av-profile': av_profile,
                'webfilter-profile': webfilter_profile,
                'ips-sensor': ips_sensor,
                'application-list': application_list,
                'ssl-ssh-profile': ssl_ssh_profile
            } %}
            {% set _utm_enable = true %}
            {% for k, v in profiles.iteritems() if v is defined and v is not none %}
               {% if _utm_enable %}
                   {%set _utm_enable = false %}
                   "utm-status": "enable",
                   "profile-protocol-options":"default",
               {% endif %}
               "{{ k }}": "{{ v }}",
            {% else %}
               "utm-status": "disable",
               "profile-protocol-options": "",
            {% endfor %}
            {% if comments is defined %}
                "comments": "{{ comments }}",
            {% endif %}
            "schedule": "always"
        }
    }
}
"""

DELETE_FIREWALL_POLICY = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall/policy/{{ id }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall/policy/{{ id }}",
    {% endif %}
    "method": "DELETE"
}
"""

GET_FIREWALL_POLICY = """
{
    {% if id is defined %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/policy/{{ id }}/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/policy/{{ id }}/",
        {% endif %}
    {% else %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/policy/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/policy/",
        {% endif %}
    {% endif %}
    "method": "GET"
}
"""


MOVE_FIREWALL_POLICY = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall/policy/{{ id }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall/policy/{{ id }}",
    {% endif %}
    "method": "PUT",
    "body": {
        {% if before is defined %}
            "before": "{{ before }}",
        {% else %}
            "after": "{{ after }}",
        {% endif %}
        "action": "move"
    }
}
"""


ADD_FIREWALL_VIP = """
{
    "path":"/api/v2/cmdb/firewall/vip/",
    "method": "POST",
    "body": {
        {% if vdom is defined %}
            "vdom": "{{ vdom }}",
        {% else %}
            "vdom": "root",
        {% endif %}
        "name": "vip",
        "json": {
            "name": "{{ name }}",
            "extip": "{{ extip }}",
            "extintf": "{{ extintf }}",
            "mappedip": [{
                    "range": "{{ mappedip }}"
            }]
        }
    }
}
"""

DELETE_FIREWALL_VIP = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall/vip/{{ name }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall/vip/{{ name }}",
    {% endif %}
    "method": "DELETE"
}
"""

GET_FIREWALL_VIP = """
{
    {% if name is defined %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/vip/{{ name }}/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/vip/{{ name }}/",
        {% endif %}
    {% else %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/vip/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/vip/",
        {% endif %}
    {% endif %}
    "method": "GET"
}
"""

ADD_FIREWALL_IPPOOL = """
{
    "path":"/api/v2/cmdb/firewall/ippool/",
    "method": "POST",
    "body": {
        {% if vdom is defined %}
            "vdom": "{{ vdom }}",
        {% else %}
            "vdom": "root",
        {% endif %}
        "name": "ippool",
        "json": {
            "startip": "{{ startip }}",
            {% if endip is defined %}
                "endip": "{{ endip }}",
            {% else %}
                "endip": "{{ startip }}",
            {% endif %}
            {% if type is defined %}
                "type": "{{ type }}",
            {% else %}
                "type": "one-to-one",
            {% endif %}
            {% if comments is defined %}
                "comments": "{{ comments }}",
            {% endif %}
            {% if name is defined %}
                "name": "{{ name }}"
            {% else %}
                "name": "{{ startip }}"
            {% endif %}
        }
    }
}
"""

DELETE_FIREWALL_IPPOOL = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall/ippool/{{ name }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall/ippool/{{ name }}/",
    {% endif %}
    "method": "DELETE"
}
"""

GET_FIREWALL_IPPOOL = """
{
    {% if name is defined %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/ippool/{{ name }}/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/ippool/{{ name }}/",
        {% endif %}
    {% else %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/ippool/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/ippool/",
        {% endif %}
    {% endif %}
    "method": "GET"
}
"""

## firewall addresses
ADD_FIREWALL_ADDRESS = """
{
    "path":"/api/v2/cmdb/firewall/address/",
    "method": "POST",
    "body": {
        {% if vdom is defined %}
            "vdom": "{{ vdom }}",
        {% else %}
            "vdom": "root",
        {% endif %}
        "name": "address",
        "json": {
            {% if associated_interface is defined %}
                "associated-interface": "{{ associated_interface }}",
            {% endif %}
            {% if comment is defined %}
                "comment": "{{ comment }}",
            {% endif %}
            "subnet": "{{ subnet }}",
            "name": "{{ name }}"
        }
    }
}
"""

SET_FIREWALL_ADDRESS = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall/address/{{ name }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall/address/{{ name }}",
    {% endif %}
    "method": "PUT",
    "body": {
        "json": {
            {% if associated_interface is defined %}
                "associated-interface": "{{ associated_interface }}",
            {% endif %}
            {% if comment is defined %}
                "comment": "{{ comment }}",
            {% endif %}
            {% if subnet is defined %}
                "subnet": "{{ subnet }}",
            {% endif %}
            "name": "{{ name }}"
        }
    }
}
"""

DELETE_FIREWALL_ADDRESS = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall/address/{{ name }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall/address/{{ name }}",
    {% endif %}
    "method": "DELETE"
}
"""

GET_FIREWALL_ADDRESS = """
{
    {% if name is defined %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/address/{{ name }}/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/address/{{ name }}/",
        {% endif %}
    {% else %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/address/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/address/",
        {% endif %}
    {% endif %}
    "method": "GET"
}
"""

## firewall address group
ADD_FIREWALL_ADDRGRP = """
{
    "path":"/api/v2/cmdb/firewall/addrgrp/",
    "method": "POST",
    "body": {
        {% if vdom is defined %}
            "vdom": "{{ vdom }}",
        {% else %}
            "vdom": "root",
        {% endif %}
        "name": "addrgrp",
        "json": {
            "name": "{{ name }}",
            "member": [
            {% for member in members[:-1] %}
                {
                    "name": "{{ member }}"
                },
            {% endfor %}
                {
                    "name": "{{ members[-1] }}"
                }
            ]
        }
    }
}
"""

SET_FIREWALL_ADDRGRP = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall/addrgrp/{{ name }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall/addrgrp/{{ name }}",
    {% endif %}
    "method": "PUT",
    "body": {
        "json": {
            "member": [
            {% for member in members[:-1] %}
                {
                    "name": "{{ member }}"
                },
            {% endfor %}
                {
                    "name": "{{ members[-1] }}"
                }
            ]
        }
    }
}
"""


DELETE_FIREWALL_ADDRGRP = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall/addrgrp/{{ name }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall/addrgrp/{{ name }}",
    {% endif %}
    "method": "DELETE"
}
"""

GET_FIREWALL_ADDRGRP = """
{
    {% if vdom is defined %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/addrgrp/{{ name }}/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/addrgrp/{{ name }}/",
        {% endif %}
    {% else %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall/addrgrp/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall/addrgrp/",
        {% endif %}
    {% endif %}
    "method": "GET"
}
"""

## firewall service custom
ADD_FIREWALL_SERVICE = """
{
    "path": "/api/v2/cmdb/firewall.service/custom/",
    "method": "POST",
    "body": {
        {% if vdom is defined %}
            "vdom": "{{ vdom }}",
        {% else %}
            "vdom": "root",
        {% endif %}
        "name": "custom",
        "json": {
            {% if protocol is defined %}
                "protocol": "{{ protocol }}",
            {% else %}
                "protocol": "TCP/UDP/SCTP",
            {% endif %}
            {% if fqdn is defined %}
                "fqdn": "{{ fqdn }}",
            {% endif %}
            {% if iprange is defined %}
                "iprange": "{{ iprange }}",
            {% endif %}
            {% if tcp_portrange is defined %}
                "tcp-portrange": "{{ tcp_portrange }}",
            {% endif %}
            {% if udp_portrange is defined %}
                "udp-portrange": "{{ udp_portrange }}",
            {% endif %}
            {% if sctp_portrange is defined %}
                "sctp-portrange": "{{ udp_portrange }}",
            {% endif %}
            {% if comment is defined %}
                "comment": "{{ comment }}",
            {% endif %}
            "name": "{{ name }}"
        }
    }
}
"""

## update firewall service custom
SET_FIREWALL_SERVICE = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall.service/custom/{{ name }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall.service/custom/{{ name }}",
    {% endif %}
    "method": "PUT",
    "body": {
        "json": {
            {% if protocol is defined %}
                "protocol": "{{ protocol }}",
            {% endif %}
            {% if fqdn is defined %}
                "fqdn": "{{ fqdn }}",
            {% endif %}
            {% if iprange is defined %}
                "iprange": "{{ iprange }}",
            {% endif %}
            {% if tcp_portrange is defined %}
                "tcp-portrange": "{{ tcp_portrange }}",
            {% else %}
                "tcp-portrange": "",
            {% endif %}
            {% if udp_portrange is defined %}
                "udp-portrange": "{{ udp_portrange }}",
            {% else %}
                "udp-portrange": "",
            {% endif %}
            {% if sctp_portrange is defined %}
                "sctp-portrange": "{{ sctp_portrange }}",
            {% else %}
                "sctp-portrange": "",
            {% endif %}
            {% if comment is defined %}
                "comment": "{{ comment }}",
            {% endif %}
            "name": "{{ name }}"
        }
    }
}
"""

DELETE_FIREWALL_SERVICE = """
{
    {% if vdom is defined %}
        "path": "/api/v2/cmdb/firewall.service/custom/{{ name }}/?vdom={{ vdom }}",
    {% else %}
        "path": "/api/v2/cmdb/firewall.service/custom/{{ name }}",
    {% endif %}
    "method": "DELETE"
}
"""

GET_FIREWALL_SERVICE = """
{
    {% if name is defined %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall.service/custom/{{ name }}/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall.service/custom/{{ name }}/",
        {% endif %}
    {% else %}
        {% if vdom is defined %}
            "path": "/api/v2/cmdb/firewall.service/custom/?vdom={{ vdom }}",
        {% else %}
            "path": "/api/v2/cmdb/firewall.service/custom/",
        {% endif %}
    {% endif %}
    "method": "GET"
}
"""



# Copyright 2015 Fortinet, Inc.
#
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import jinja2
from oslo_log import log as logging
from oslo_serialization import jsonutils

from networking_fortinet._i18n import _LE
from networking_fortinet.api_client import base
from networking_fortinet.api_client import eventlet_client
from networking_fortinet.api_client import eventlet_request
from networking_fortinet.api_client import exception
from networking_fortinet.api_client import templates
from networking_fortinet.common import singleton

LOG = logging.getLogger(__name__)


@singleton.singleton
class FortiosApiClient(eventlet_client.EventletApiClient):
    """The FortiOS API Client."""

    def __init__(self, api_providers, user, password,
                 concurrent_connections=base.DEFAULT_CONCURRENT_CONNECTIONS,
                 gen_timeout=base.GENERATION_ID_TIMEOUT,
                 use_https=False,
                 connect_timeout=base.DEFAULT_CONNECT_TIMEOUT,
                 http_timeout=75, retries=2, redirects=2):
        '''Constructor. Adds the following:
        :param api_providers: a list of tuples of the form: (host, port,
            is_ssl)
        :param http_timeout: how long to wait before aborting an
            unresponsive controller (and allow for retries to another
            controller in the cluster)
        :param retries: the number of concurrent connections.
        :param redirects: the number of concurrent connections.
        '''
        super(FortiosApiClient, self).__init__(
            api_providers, user, password,
            concurrent_connections=concurrent_connections,
            gen_timeout=gen_timeout, use_https=use_https,
            connect_timeout=connect_timeout)

        self._request_timeout = http_timeout * retries
        self._http_timeout = http_timeout
        self._retries = retries
        self._redirects = redirects
        self._version = None
        self.message = {}
        self._user = user
        self._password = password

    @staticmethod
    def _render(template, **message):
        '''Render API message from it's template

        :param template: defined API message with essential params.
        :param message: It is a dictionary, included values of the params
                        for the template
        '''
        if not message:
            message = {}
        msg = jinja2.Template(template).render(**message)
        return jsonutils.loads(msg)

    def request(self, opt, content_type="application/json", **message):
        '''Issues request to controller.'''
        self.message = self._render(getattr(templates, opt), **message)
        method = self.message['method']
        url = self.message['path']
        body = self.message['body'] if 'body' in self.message else None
        g = eventlet_request.GenericRequestEventlet(
            self, method, url, body, content_type, auto_login=True,
            http_timeout=self._http_timeout,
            retries=self._retries, redirects=self._redirects)
        g.start()
        response = g.join()

        # response is a modified HTTPResponse object or None.
        # response.read() will not work on response as the underlying library
        # request_eventlet.ApiRequestEventlet has already called this
        # method in order to extract the body and headers for processing.
        # ApiRequestEventlet derived classes call .read() and
        # .getheaders() on the HTTPResponse objects and store the results in
        # the response object's .body and .headers data members for future
        # access.

        if response is None:
            # Timeout.
            LOG.error(_LE('Request timed out: %(method)s to %(url)s'),
                      {'method': method, 'url': url})
            raise exception.RequestTimeout()

        status = response.status
        if status == 401:
            raise exception.UnAuthorizedRequest()
        # Fail-fast: Check for exception conditions and raise the
        # appropriate exceptions for known error codes.
        if status in exception.ERROR_MAPPINGS:
            LOG.error(_LE("Received error code: %s"), status)
            LOG.error(_LE("Server Error Message: %s"), response.body)
            exception.ERROR_MAPPINGS[status](response)

        # Continue processing for non-error condition.
        if (status != 200 and status != 201
                and status != 204):
            LOG.error(_LE("%(method)s to %(url)s, unexpected response code: "
                        "%(status)d (content = '%(body)s')"),
                      {'method': method, 'url': url,
                       'status': response.status, 'body': response.body})
            return None

        if url == jsonutils.loads(templates.LOGOUT)['path']:
            return response.body
        else:
            try:
                return jsonutils.loads(response.body)
            except UnicodeDecodeError:
                LOG.debug("The following strings cannot be decoded with "
                          "'utf-8, trying 'ISO-8859-1' instead. %(body)s",
                          {'body': response.body})
                return jsonutils.loads(response.body, encoding='ISO-8859-1')
            except Exception as e:
                LOG.error(_LE("Decode error, the response.body %(body)s"),
                          {'body': response.body})
                raise e



# Copyright 2015 Fortinet, Inc.
#
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#

import time

import eventlet
eventlet.monkey_patch()
from oslo_log import log as logging

from networking_fortinet.api_client import base
from networking_fortinet.api_client import eventlet_request

LOG = logging.getLogger(__name__)


class EventletApiClient(base.ApiClientBase):
    """Eventlet-based implementation of FortiOS ApiClient ABC."""

    def __init__(self, api_providers, user, password,
                 concurrent_connections=base.DEFAULT_CONCURRENT_CONNECTIONS,
                 gen_timeout=base.GENERATION_ID_TIMEOUT,
                 use_https=True,
                 connect_timeout=base.DEFAULT_CONNECT_TIMEOUT):
        '''Constructor

        :param api_providers: a list of tuples of the form: (host, port,
            is_ssl).
        :param user: login username.
        :param password: login password.
        :param concurrent_connections: total number of concurrent connections.
        :param use_https: whether or not to use https for requests.
        :param connect_timeout: connection timeout in seconds.
        :param gen_timeout controls how long the generation id is kept
            if set to -1 the generation id is never timed out
        '''
        if not api_providers:
            api_providers = []
        self._api_providers = set([tuple(p) for p in api_providers])
        self._api_provider_data = {}  # tuple(semaphore, session_cookie)
        for p in self._api_providers:
            self._set_provider_data(p, (eventlet.semaphore.Semaphore(1), None))
        self._user = user
        self._password = password
        self._concurrent_connections = concurrent_connections
        self._use_https = use_https
        self._connect_timeout = connect_timeout
        self._config_gen = None
        self._config_gen_ts = None
        self._gen_timeout = gen_timeout

        # Connection pool is a list of queues.
        self._conn_pool = eventlet.queue.PriorityQueue()
        self._next_conn_priority = 1
        for host, port, is_ssl in api_providers:
            for _ in range(concurrent_connections):
                conn = self._create_connection(host, port, is_ssl)
                self._conn_pool.put((self._next_conn_priority, conn))
                self._next_conn_priority += 1

    def acquire_redirect_connection(self, conn_params, auto_login=True,
                                    headers=None):
        """Check out or create connection to redirected NSX API server.

        Args:
            conn_params: tuple specifying target of redirect, see
                self._conn_params()
            auto_login: returned connection should have valid session cookie
            headers: headers to pass on if auto_login

        Returns: An available HTTPConnection instance corresponding to the
                 specified conn_params. If a connection did not previously
                 exist, new connections are created with the highest prioity
                 in the connection pool and one of these new connections
                 returned.
        """
        result_conn = None
        data = self._get_provider_data(conn_params)
        if data:
            # redirect target already exists in provider data and connections
            # to the provider have been added to the connection pool. Try to
            # obtain a connection from the pool, note that it's possible that
            # all connection to the provider are currently in use.
            conns = []
            while not self._conn_pool.empty():
                priority, conn = self._conn_pool.get_nowait()
                if not result_conn and self._conn_params(conn) == conn_params:
                    conn.priority = priority
                    result_conn = conn
                else:
                    conns.append((priority, conn))
            for priority, conn in conns:
                self._conn_pool.put((priority, conn))
            # hack: if no free connections available, create new connection
            # and stash "no_release" attribute (so that we only exceed
            # self._concurrent_connections temporarily)
            if not result_conn:
                conn = self._create_connection(*conn_params)
                conn.priority = 0  # redirect connections have highest priority
                conn.no_release = True
                result_conn = conn
        else:
            #redirect target not already known, setup provider lists
            self._api_providers.update([conn_params])
            self._set_provider_data(conn_params,
                                    (eventlet.semaphore.Semaphore(1), None))
            # redirects occur during cluster upgrades, i.e. results to old
            # redirects to new, so give redirect targets highest priority
            priority = 0
            for i in range(self._concurrent_connections):
                conn = self._create_connection(*conn_params)
                conn.priority = priority
                if i == self._concurrent_connections - 1:
                    break
                self._conn_pool.put((priority, conn))
            result_conn = conn
        if result_conn:
            result_conn.last_used = time.time()
            if auto_login and self.auth_cookie(conn) is None:
                self._wait_for_login(result_conn, headers)
        return result_conn

    def _login(self, conn=None, headers=None):
        '''Issue login request and update authentication cookie.'''
        cookie = None
        g = eventlet_request.LoginRequestEventlet(
            self, self._user, self._password, conn, headers)
        g.start()
        ret = g.join()
        if ret:
            if isinstance(ret, Exception):
                LOG.error(_('Login error "%s"'), ret)
                raise ret

            cookie = ret.getheader("Set-Cookie")
            if cookie:
                LOG.debug("Saving new authentication cookie '%s'", cookie)

        return cookie

# Register as subclass.
base.ApiClientBase.register(EventletApiClient)



# Copyright 2015 Fortinet, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import unittest2

from networking_fortinet.api_client import client
from networking_fortinet.tests.common import utils

PREFIX = 'test_CRUD_'


class ClientTestCase(unittest2.TestCase):
    def setUp(self, fgtip=None, port=443, protocol='https', username='admin',
              password=''):
        super(ClientTestCase, self).setUp()
        fgtip = '10.160.37.96' if not fgtip else fgtip
        api_server = [(fgtip, port, 'https' == protocol)]
        self.client = client.FortiosApiClient(api_server, username, password)
        self.vdom = 'root'

    def tearDown(self):
        self.client.request('LOGOUT')
        super(ClientTestCase, self).tearDown()

    def _func_name(self):
        return inspect.stack()[1][3]

    def _test(self, name, create_msg, query_msg, update_msg, key=None):
        if name.startswith(PREFIX):
            name = name[len(PREFIX):]
        name = name.upper()
        with utils.Prepare_vdom(self.client, vdom=self.vdom) as env:
            try:
                create_msg.setdefault('vdom', env.vdom)
                res = self.client.request('ADD_%s' % name, **create_msg)
                self.assertEqual(200, res.get('http_status', None))
                if key:
                    query_msg.setdefault('vdom', env.vdom)
                    query_msg.setdefault(
                        key, res['results'].get('mkey', None))
                    update_msg.setdefault('vdom', env.vdom)
                    update_msg.setdefault(
                        key, res['results'].get('mkey', None))

                res = self.client.request('GET_%s' % name, **query_msg)
                self.assertEqual(200, res.get('http_status', None))
                if update_msg:
                    res = self.client.request('SET_%s' % name, **update_msg)
                    self.assertEqual(200, res.get('http_status', None))
                    res = self.client.request('GET_%s' % name, **query_msg)
                    self.assertEqual(200, res.get('http_status', None))

                res = self.client.request('DELETE_%s' % name, **query_msg)
                self.assertEqual(200, res.get('http_status', None))
            except Exception:
                self.client.request('DELETE_%s' % name, **query_msg)
                #print("No %(name)s need to be deleted" % {'name': name})

    def test_CRUD_firewall_policy(self):
        create_msg = {
            'dstintf': 'port3',
            'srcintf': 'port4',
            'action': 'deny',
            'webfilter_profile': 'default'
        }
        query_msg = {}
        update_msg = {
            'dstintf': 'port5',
            'srcintf': 'port6',
            'action': 'deny'
        }
        self._test(
            self._func_name(), create_msg, query_msg, update_msg, key='id')

    def test_CRUD_firewall_address(self):
        name = self._func_name()
        create_msg = {
            'name': name,
            "subnet": "192.168.44.0 255.255.255.0"
        }
        query_msg = {'name': name}
        update_msg = {
            'name': name,
            "subnet": "10.1.1.0 255.255.255.252"
        }
        self._test(name, create_msg, query_msg, update_msg)

    def test_CRUD_firewall_service(self):
        create_msg = {
            'name': self._func_name(),
            'tcp_portrange': '100-200:300-400',
            'comment': self._func_name()
        }
        query_msg = {'name': self._func_name()}
        update_msg = {
            'name': self._func_name(),
            'protocol': 'ICMP'
        }
        self._test(self._func_name(), create_msg, query_msg, update_msg)


if __name__ == '__main__':
    unittest2.main()



# Copyright 2015 Fortinet Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
from netaddr import IPAddress
from netaddr import IPNetwork
from oslo_log import log as logging
from oslo_utils import excutils

from neutron.api import extensions as neutron_extensions
from neutron.api.v2 import attributes as attr
from neutron.common import constants as l3_consts
from neutron import context as neutron_context
from neutron.db import l3_db
from neutron import manager
from neutron.plugins.common import constants as const
from neutron_fwaas.db.firewall import firewall_db
from neutron_fwaas.db.firewall import firewall_router_insertion_db
import neutron_fwaas.extensions as extensions
from neutron_fwaas.extensions import firewall as fw_ext

from networking_fortinet._i18n import _LE
from networking_fortinet.common import config
from networking_fortinet.common import constants as constants
from networking_fortinet.common import utils as utils
from networking_fortinet.db import models as fortinet_db
from networking_fortinet.tasks import constants as t_consts
from networking_fortinet.tasks import tasks

LOG = logging.getLogger(__name__)
FORTINET_FW = "fortinet_fw"
FORTINET_FW_PLUGIN = "fortinet_fw_plugin"


class FortinetFirewallPlugin(
    firewall_db.Firewall_db_mixin,
    firewall_router_insertion_db.FirewallRouterInsertionDbMixin):
    """Implementation of the Neutron Firewall Service Plugin.

    This class manages the workflow of FWaaS request/response.
    Most DB related works are implemented in class
    firewall_db.Firewall_db_mixin.
    """
    neutron_extensions.append_api_extensions_path(extensions.__path__)
    supported_extension_aliases = ["fwaas", "fwaasrouterinsertion"]
    path_prefix = fw_ext.FIREWALL_PREFIX

    def __init__(self):
        """Do the initialization for the firewall service plugin here."""
        self._fortigate = config.fgt_info
        self._driver = config.get_apiclient()
        self.task_manager = tasks.TaskManager()
        self.task_manager.start()
        firewall_db.subscribe()

    def _rpc_update_firewall(self, context, firewall_id):
        status_update = {"firewall": {"status": const.PENDING_UPDATE}}
        super(FortinetFirewallPlugin, self).update_firewall(
            context, firewall_id, status_update)
        fw_with_rules = self._make_firewall_dict_with_rules(context,
                                                            firewall_id)
        # this is triggered on an update to fw rule or policy, no
        # change in associated routers.
        fw_with_rules['add-router-ids'] = self.get_firewall_routers(
                context, firewall_id)
        fw_with_rules['del-router-ids'] = []
        self._apply_firewall(context, **fw_with_rules)

    def _rpc_update_firewall_policy(self, context, firewall_policy_id):
        firewall_policy = self.get_firewall_policy(context, firewall_policy_id)
        if firewall_policy:
            for firewall_id in firewall_policy['firewall_list']:
                self._rpc_update_firewall(context, firewall_id)

    def _ensure_update_firewall(self, context, firewall_id):
        fwall = self.get_firewall(context, firewall_id)
        if fwall['status'] in [const.PENDING_CREATE,
                               const.PENDING_UPDATE,
                               const.PENDING_DELETE]:
            raise fw_ext.FirewallInPendingState(firewall_id=firewall_id,
                                                pending_state=fwall['status'])

    def _ensure_update_firewall_policy(self, context, firewall_policy_id):
        firewall_policy = self.get_firewall_policy(context, firewall_policy_id)
        if firewall_policy and 'firewall_list' in firewall_policy:
            for firewall_id in firewall_policy['firewall_list']:
                self._ensure_update_firewall(context, firewall_id)

    def _ensure_update_firewall_rule(self, context, firewall_rule_id):
        fw_rule = self.get_firewall_rule(context, firewall_rule_id)
        if 'firewall_policy_id' in fw_rule and fw_rule['firewall_policy_id']:
            self._ensure_update_firewall_policy(context,
                                                fw_rule['firewall_policy_id'])

    def _get_routers_for_create_firewall(self, tenant_id, context, firewall):
        # pop router_id as this goes in the router association db
        # and not firewall db
        LOG.debug("# _get_routers_for_create_firewall called Fortinet_plugin")
        router_ids = firewall['firewall'].pop('router_ids', None)
        if router_ids == attr.ATTR_NOT_SPECIFIED:
            # old semantics router-ids keyword not specified pick up
            # all routers on tenant.
            l3_plugin = manager.NeutronManager.get_service_plugins().get(
                const.L3_ROUTER_NAT)
            ctx = neutron_context.get_admin_context()
            routers = l3_plugin.get_routers(ctx)
            router_ids = [
                router['id']
                for router in routers
                if router['tenant_id'] == tenant_id]
            # validation can still fail this if there is another fw
            # which is associated with one of these routers.
            self.validate_firewall_routers_not_in_use(context, router_ids)
            return router_ids
        else:
            if not router_ids:
                # This indicates that user specifies no routers.
                return []
            else:
                # some router(s) provided.
                self.validate_firewall_routers_not_in_use(context, router_ids)
                return router_ids

    def create_firewall(self, context, firewall):
        LOG.debug("create_firewall() called Fortinet_plugin")
        tenant_id = firewall['firewall']['tenant_id']
        fw_new_rtrs = self._get_routers_for_create_firewall(
            tenant_id, context, firewall)
        if not fw_new_rtrs:
            # no messaging to agent needed, and fw needs to go
            # to INACTIVE(no associated rtrs) state.
            status = const.INACTIVE
            fw = super(FortinetFirewallPlugin, self).create_firewall(
                context, firewall, status)
            fw['router_ids'] = []
            return fw
        else:
            fw = super(FortinetFirewallPlugin, self).create_firewall(
                context, firewall)
            fw['router_ids'] = fw_new_rtrs

        fw_with_rules = (
            self._make_firewall_dict_with_rules(context, fw['id']))

        fw_with_rtrs = {'fw_id': fw['id'], 'router_ids': fw_new_rtrs}
        self.set_routers_for_firewall(context, fw_with_rtrs)
        fw_with_rules['add-router-ids'] = fw_new_rtrs
        fw_with_rules['del-router-ids'] = []

        self._apply_firewall(context, **fw_with_rules)
        return fw

    def update_firewall(self, context, id, firewall):
        LOG.debug("Fortinet_plugin update_firewall() called, "
                  "id is %(id)s, firewall is %(fw)s",
                  {'id': id, 'fw': firewall})
        self._ensure_update_firewall(context, id)
        # pop router_id as this goes in the router association db
        # and not firewall db
        router_ids = firewall['firewall'].pop('router_ids', None)
        fw_current_rtrs = self.get_firewall_routers(context, id)
        if router_ids is not None:
            if router_ids == []:
                # This indicates that user is indicating no routers.
                fw_new_rtrs = []
            else:
                self.validate_firewall_routers_not_in_use(
                    context, router_ids, id)
                fw_new_rtrs = router_ids
            self.update_firewall_routers(context, {'fw_id': id,
                'router_ids': fw_new_rtrs})
        else:
            # router-ids keyword not specified for update pick up
            # existing routers.
            fw_new_rtrs = self.get_firewall_routers(context, id)

        if not fw_new_rtrs and not fw_current_rtrs:
            # no messaging to agent needed, and we need to continue
            # in INACTIVE state
            firewall['firewall']['status'] = const.INACTIVE
            fw = super(FortinetFirewallPlugin, self).update_firewall(
                context, id, firewall)
            fw['router_ids'] = []
            return fw
        else:
            firewall['firewall']['status'] = const.PENDING_UPDATE
            fw = super(FortinetFirewallPlugin, self).update_firewall(
                context, id, firewall)
            fw['router_ids'] = fw_new_rtrs

        fw_with_rules = (
            self._make_firewall_dict_with_rules(context, fw['id']))

        # determine rtrs to add fw to and del from
        fw_with_rules['add-router-ids'] = fw_new_rtrs
        fw_with_rules['del-router-ids'] = list(
            set(fw_current_rtrs).difference(set(fw_new_rtrs)))

        # last-router drives agent to ack with status to set state to INACTIVE
        fw_with_rules['last-router'] = not fw_new_rtrs

        LOG.debug("## update_firewall %s: Add Routers: %s, Del Routers: %s",
            fw['id'],
            fw_with_rules['add-router-ids'],
            fw_with_rules['del-router-ids'])
        self._apply_firewall(context, **fw_with_rules)
        #self.agent_rpc.update_firewall(context, fw_with_rules)
        return fw

    def update_firewall_for_delete_router(self, context, router_id):
        LOG.debug("fwaas delete_router() called, router_id: %(rtid)s",
                  {'rtid': router_id})
        cls = firewall_router_insertion_db.FirewallRouterAssociation
        db_fw_rt = fortinet_db.query_record(context, cls, router_id=router_id)
        if not db_fw_rt:
            return None
        firewall = {u'firewall': {'router_ids': []}}
        return self.update_firewall(context, db_fw_rt.fw_id, firewall)

    def delete_db_firewall_object(self, context, id):
        super(FortinetFirewallPlugin, self).delete_firewall(context, id)

    def delete_firewall(self, context, id):
        LOG.debug("Fortinet_plugin delete_firewall() called, fw_id %(id)s",
                  {'id': id})
        fw_with_rules = (
            self._make_firewall_dict_with_rules(context, id))

        status = {"firewall": {"status": const.PENDING_DELETE}}
        super(FortinetFirewallPlugin, self).update_firewall(
            context, id, status)
        # Reflect state change in fw_with_rules
        fw_with_rules['del-router-ids'] = self.get_firewall_routers(
            context, id)
        self._apply_firewall(context, **fw_with_rules)
        self.delete_db_firewall_object(context, id)

    def update_firewall_policy(self, context, id, firewall_policy):
        LOG.debug("update_firewall_policy called, "
                  "id =%(id)s, firewall_policy=%(fp)s",
                  {'id': id, 'fp': firewall_policy})
        self._ensure_update_firewall_policy(context, id)
        firewall_policy_old = self.get_firewall_policy(context, id)
        firewall_rule_ids = firewall_policy_old.get('firewall_rules', [])
        tenant_id = firewall_policy_old.get('tenant_id', None)
        fwp = super(FortinetFirewallPlugin,
                    self).update_firewall_policy(context, id, firewall_policy)
        for fwr_id in firewall_rule_ids:
            fw_rule = self.get_firewall_rule(context, fwr_id)
            self._delete_firewall_rule(context, tenant_id, **fw_rule)
        self._rpc_update_firewall_policy(context, id)
        return fwp

    def create_firewall_rule(self, context, firewall_rule):
        """
        :param context:
        :param firewall_rule:
        firewall_rule={'firewall_rule': {... }}
        :return:
        """
        LOG.debug("create_firewall_rule() firewall_rule=%(fwr)s",
                  {'fwr': firewall_rule})
        return super(FortinetFirewallPlugin,
            self).create_firewall_rule(context, firewall_rule)

    def delete_firewall_rule(self, context, id):
        super(FortinetFirewallPlugin, self).delete_firewall_rule(context, id)

    def update_firewall_rule(self, context, id, firewall_rule):
        LOG.debug("update_firewall_rule() id: %(id)s, "
                  "firewall_rule: %(firewall_rule)s",
                  {'id': id, 'firewall_rule': firewall_rule})
        try:
            fwr = self._update_firewall_rule_dict(context, id, firewall_rule)
            self._update_firewall_rule(context, id, fwr)
            self._ensure_update_firewall_rule(context, id)
            fwr = super(FortinetFirewallPlugin,
                        self).update_firewall_rule(context, id, firewall_rule)
            utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)
            return fwr
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("update_firewall_rule %(fwr)s failed"),
                          {'fwr': firewall_rule})
                utils._rollback_on_err(self, context, e)

    def insert_rule(self, context, id, rule_info):
        self._ensure_update_firewall_policy(context, id)
        try:
            fwp = super(FortinetFirewallPlugin,
                    self).insert_rule(context, id, rule_info)
            self._rpc_update_firewall_policy(context, id)
            utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)
            return fwp
        except Exception as e:
            with excutils.save_and_reraise_exception():
                self.remove_rule(context, id, rule_info)
                utils._rollback_on_err(self, context, e)

    def remove_rule(self, context, id, rule_info):
        LOG.debug("Fortinet_plugin remove_rule() called")
        self._ensure_update_firewall_policy(context, id)
        if rule_info.get('firewall_rule_id', None):
            firewall_rule = self._get_firewall_rule(
                context, rule_info['firewall_rule_id'])
            fwr = self._make_firewall_rule_dict(firewall_rule)
            self._delete_firewall_rule(context, fwr['tenant_id'], **fwr)
        fwp = super(FortinetFirewallPlugin, self).remove_rule(
            context, id, rule_info)
        self._rpc_update_firewall_policy(context, id)
        return fwp

    def get_firewalls(self, context, filters=None, fields=None):
        LOG.debug("fwaas get_firewalls() called, filters=%(filters)s, "
                  "fields=%(fields)s",
                  {'filters': filters, 'fields': fields})
        fw_list = super(FortinetFirewallPlugin, self).get_firewalls(
                        context, filters, fields)
        for fw in fw_list:
            fw_current_rtrs = self.get_firewall_routers(context, fw['id'])
            fw['router_ids'] = fw_current_rtrs
        return fw_list

    def get_firewall(self, context, id, fields=None):
        LOG.debug("fwaas get_firewall() called")
        res = super(FortinetFirewallPlugin, self).get_firewall(
                        context, id, fields)
        fw_current_rtrs = self.get_firewall_routers(context, id)
        res['router_ids'] = fw_current_rtrs
        return res

    def _apply_firewall(self, context, **fw_with_rules):
        tenant_id = fw_with_rules['tenant_id']
        default_fwr = self._make_default_firewall_rule_dict(tenant_id)
        try:
            if fw_with_rules.get('del-router-ids', None):
                for fwr in list(fw_with_rules.get('firewall_rule_list', None)):
                    self._delete_firewall_rule(context, tenant_id, **fwr)
                if default_fwr:
                    self._delete_firewall_rule(
                        context, tenant_id, **default_fwr)
                self.update_firewall_status(
                    context, fw_with_rules['id'], const.INACTIVE)

            if fw_with_rules.get('add-router-ids', None):
                vdom = getattr(
                    fortinet_db.Fortinet_ML2_Namespace.query_one(
                        context, tenant_id=tenant_id), 'vdom', None)
                if not vdom:
                    raise fw_ext.FirewallInternalDriverError(
                        driver='Fortinet_fwaas_plugin')
                if default_fwr:
                    self._add_firewall_rule(context, tenant_id, **default_fwr)
                for fwr in reversed(
                        list(fw_with_rules.get('firewall_rule_list', None))):
                    self._add_firewall_rule(context, tenant_id, **fwr)
                self.update_firewall_status(
                    context, fw_with_rules['id'], const.ACTIVE)
            else:
                self.update_firewall_status(
                    context, fw_with_rules['id'], const.INACTIVE)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("apply_firewall %(fws)s failed"),
                          {'fws': fw_with_rules})
                utils._rollback_on_err(self, context, e)
        utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)

    def _add_firewall_rule(self, context, fwp_tenant_id, **fwr):
        """
        :param obj:
        :param context:
        :param kwargs: dictionary, firewall rule
        firewall_rule: {'source_ip_address': u'192.176.10.0/24',... }
        :return:
        """
        LOG.debug("# _add_firewall_rule() called")
        namespace = fortinet_db.Fortinet_ML2_Namespace.query_one(
            context, tenant_id=fwp_tenant_id)
        vdom = getattr(namespace, 'vdom', None)
        if not vdom or not fwr:
            return None
        inf_int, inf_ext = utils.get_vlink_intf(
            self, context, vdom=namespace.vdom)
        srcaddr = self._add_fwr_ip_address(
            context, vdom, place='source_ip_address', **fwr)
        dstaddr = self._add_fwr_ip_address(
            context, vdom, place='destination_ip_address', **fwr)
        service = self._add_fwr_service(context, vdom, **fwr)
        action = self._get_fwr_action(**fwr)
        profiles = self._get_fwp_profiles(action)
        match_vip = 'enable'
        name = fwr.get('name', '')
        # add a basic firewall rule('accept': incoming, 'deny': bidirectional)
        fortinet_fwp = utils.add_fwpolicy_to_head(self, context,
                                                  vdom=vdom,
                                                  srcaddr=srcaddr['name'],
                                                  srcintf='any',
                                                  dstaddr=dstaddr['name'],
                                                  dstintf='any',
                                                  service=service['name'],
                                                  match_vip=match_vip,
                                                  action=action,
                                                  comments=name,
                                                  **profiles)
        utils.add_record(self, context,
                         fortinet_db.Fortinet_FW_Rule_Association,
                         fwr_id=fwr['id'],
                         fortinet_pid=fortinet_fwp.id,
                         type=constants.TYPE_INT)

        if action in ['accept']:
            # if allow, for the outgoing traffic it need to enable nat
            fortinet_fwp = utils.add_fwpolicy_to_head(self, context,
                                                      vdom=vdom,
                                                      srcaddr=srcaddr['name'],
                                                      srcintf='any',
                                                      dstaddr=dstaddr['name'],
                                                      dstintf=inf_int,
                                                      nat='enable',
                                                      service=service['name'],
                                                      action=action,
                                                      comments=name,
                                                      **profiles)
            utils.add_record(self, context,
                             fortinet_db.Fortinet_FW_Rule_Association,
                             fwr_id=fwr['id'],
                             fortinet_pid=fortinet_fwp.id,
                             type=constants.TYPE_EXT)

    def _update_firewall_rule(self, context, id, firewall_rule):
        LOG.debug("# _add_firewall_rule() called")
        fwps_int = fortinet_db.Fortinet_FW_Rule_Association.query_all(
            context, fwr_id=id, type=constants.TYPE_INT)
        fwps_ext = fortinet_db.Fortinet_FW_Rule_Association.query_all(
            context, fwr_id=id, type=constants.TYPE_EXT)
        if fwps_ext and fwps_int:
            fwps = fwps_int + fwps_ext
        else:
            fwps = fwps_int or fwps_ext
        if not fwps:
            return
        firewall_rule.setdefault('id', id)
        srcaddr = self._make_fortinet_fwaddress_dict(
            place='source_ip_address', **firewall_rule)
        dstaddr = self._make_fortinet_fwaddress_dict(
            place='destination_ip_address', **firewall_rule)
        service = self._make_fortinet_fwservice_dict(**firewall_rule)
        action = self._get_fwr_action(**firewall_rule)
        profiles = self._get_fwp_profiles(action)
        for fwp in fwps_int:
            vdom = fwp.fortinet_policy.vdom
            if service['name'] != 'ALL':
                utils.set_fwservice(self, context, vdom=vdom, **service)
            if srcaddr['name'] != 'all':
                utils.set_fwaddress(self, context, vdom=vdom, **srcaddr)
            if dstaddr['name'] != 'all':
                utils.set_fwaddress(self, context, vdom=vdom, **dstaddr)
        # check whether related firewall policies need to update
        fwp = fwps_int[0].fortinet_policy
        name = firewall_rule.setdefault('name', fwp.comments)
        if fwp.srcaddr == srcaddr['name'] and fwp.action == action and \
            fwp.dstaddr == dstaddr['name'] and fwp.service == service['name']:
            return
        if action in ['accept']:
            for fwp in fwps:
                fortinet_fwp = utils.set_fwpolicy(self, context,
                                                  id=fwp.fortinet_pid,
                                                  srcaddr=srcaddr['name'],
                                                  dstaddr=dstaddr['name'],
                                                  service=service['name'],
                                                  action=action,
                                                  comments=name,
                                                  **profiles)
                if not fwps_ext:
                    inf_int, inf_ext = utils.get_vlink_intf(
                        self, context, vdom=fortinet_fwp.vdom)
                    utils.add_fwaas_subpolicy(self, context,
                                              before=fortinet_fwp.edit_id,
                                              vdom=fortinet_fwp.vdom,
                                              srcaddr=srcaddr['name'],
                                              dstaddr=dstaddr['name'],
                                              dstintf=inf_int,
                                              nat='enable',
                                              service=service['name'],
                                              action=action,
                                              comments=name,
                                              fwr_id=id,
                                              type=constants.TYPE_EXT,
                                              **profiles)
        elif action in ['deny']:
            for fwp_ext in fwps_ext:
                utils.delete_fwaas_subpolicy(self, context,
                                             fwr_id=fwp_ext.fwr_id,
                                             fortinet_pid=fwp_ext.fortinet_pid)
            for fwp in fwps_int:
                utils.set_fwpolicy(self, context,
                                   id=fwp.fortinet_pid,
                                   srcaddr=srcaddr['name'],
                                   dstaddr=dstaddr['name'],
                                   service=service['name'],
                                   action=action,
                                   comments=name,
                                   **profiles)
        for fwp in fwps_int:
            vdom = fwp.fortinet_policy.vdom
            if service['name'] == 'ALL':
                #delete all old services if exist
                utils.delete_fwservice(self, context, vdom=vdom, name=id)
            if srcaddr['name'] == 'all':
                name = constants.PREFIX['source_ip_address'] + id
                utils.delete_fwaddress(self, context, vdom=vdom, name=name)
            if dstaddr['name'] == 'all':
                name = constants.PREFIX['destination_ip_address'] + id
                utils.delete_fwaddress(self, context, vdom=vdom, name=name)

    def _delete_firewall_rule(self, context, fwp_tenant_id, **fwr):
        """
        :param obj:
        :param context:
        :param kwargs: dictionary, firewall rule
        firewall_rule: {'source_ip_address': u'192.176.10.0/24',... }
        :return:
        """
        # need to consider shared firewall rules
        LOG.debug("# _delete_firewall_rule() called")
        namespace = fortinet_db.Fortinet_ML2_Namespace.query_one(
            context, tenant_id=fwp_tenant_id)
        if not namespace:
            return None
        fwp_assed = fortinet_db.Fortinet_FW_Rule_Association.query_all(
            context, fwr_id=fwr['id'])
        for fwp in fwp_assed:
            fortinet_db.delete_record(
                context, fortinet_db.Fortinet_FW_Rule_Association,
                fwr_id=fwp.fwr_id, fortinet_pid=fwp.fortinet_pid)
            utils.delete_fwpolicy(
                self, context, id=fwp.fortinet_pid, vdom=namespace.vdom)

        if fwr.get('source_ip_address', None):
            srcaddr = constants.PREFIX['source_ip_address'] + fwr['id']
            utils.delete_fwaddress(
                self, context, vdom=namespace.vdom, name=srcaddr)
        if fwr.get('destination_ip_address', None):
            dstaddr = constants.PREFIX['destination_ip_address'] + fwr['id']
            utils.delete_fwaddress(
                self, context, vdom=namespace.vdom, name=dstaddr)
        self._delete_fwr_service(context, namespace.vdom, **fwr)

    def _update_firewall_rule_dict(self, context, id, firewall_rule):
        fwr = firewall_rule.get('firewall_rule', firewall_rule)
        fwr_db = self._get_firewall_rule(context, id)
        if fwr_db.firewall_policy_id:
            fwp_db = self._get_firewall_policy(context,
                                               fwr_db.firewall_policy_id)
            if 'shared' in fwr and not fwr['shared']:
                if fwr_db['tenant_id'] != fwp_db['tenant_id']:
                    raise fw_ext.FirewallRuleInUse(firewall_rule_id=id)
        fwr.setdefault('source_port',
                       self._get_port_range_from_min_max_ports(
                           fwr_db['source_port_range_min'],
                           fwr_db['source_port_range_max']))
        fwr.setdefault('destination_port',
                       self._get_port_range_from_min_max_ports(
                           fwr_db['destination_port_range_min'],
                           fwr_db['destination_port_range_max']))
        keys = ['name', 'protocol', 'action', 'shared', 'ip_version',
                'source_ip_address', 'destination_ip_address', 'enabled']
        for key in keys:
            fwr.setdefault(key, fwr_db[key])
        return fwr

    def _make_default_firewall_rule_dict(self, tenant_id):
        if tenant_id and self._fortigate["enable_default_fwrule"]:
            return {'id': tenant_id,
                    'tenant_id': tenant_id,
                    'name': '_default_rule_deny_all',
                    'description': '_default_rule_deny_all',
                    'protocol': None,
                    'source_ip_address': None,
                    'source_port': None,
                    'destination_ip_address': None,
                    'destination_port': None,
                    'action': 'deny'}
        else:
            return {}

    def _add_fwr_ip_address(self, context, vdom,
                            place='source_ip_address', **fwr):
        fwaddress = self._make_fortinet_fwaddress_dict(
            place=place, vdom=vdom, **fwr)
        utils.add_fwaddress(self, context, **fwaddress)
        return fwaddress

    def _make_fortinet_fwaddress_dict(self, place='source_ip_address',
                                      vdom=None, **fwr):
        fwaddress = {}
        if place not in ['source_ip_address', 'destination_ip_address']:
            raise ValueError("_add_fwr_ip_address() value error of where")
        if vdom:
            fwaddress.setdefault('vdom', vdom)
        if fwr.get(place, None):
            fwaddress.setdefault('name', constants.PREFIX[place] + fwr['id'])
            fwaddress.setdefault('subnet', utils.get_subnet(fwr[place]))
        else:
            fwaddress.setdefault('name', 'all')
        return fwaddress

    def _add_fwr_service(self, context, vdom, **fwr):
        kw_service = self._make_fortinet_fwservice_dict(vdom=vdom, **fwr)
        utils.add_fwservice(self, context, **kw_service)
        return kw_service

    def _make_fortinet_fwservice_dict(self, vdom=None, **fwr):
        LOG.debug("_make_fortinet_fwservice_dict() fwr=%(fwr)s", {'fwr': fwr})
        kw_service = {}
        if vdom:
            kw_service.setdefault('vdom', vdom)
        if fwr['protocol'] in ['any', None] and \
            not fwr['destination_port'] and not fwr['source_port']:
            # SamSu: The firewall service 'all' was already added by default
            kw_service.setdefault('name', 'ALL')
        else:
            portrange = ':'.join([
                utils.port_range(fwr['destination_port']),
                utils.port_range(fwr['source_port'])])
            if fwr['protocol'] in ['tcp']:
                kw_service.setdefault('tcp_portrange', portrange)
            elif fwr['protocol'] in ['udp']:
                kw_service.setdefault('udp_portrange', portrange)
            elif fwr['protocol'] in ['icmp']:
                kw_service.setdefault('protocol', 'ICMP')
            kw_service.setdefault('name', fwr['id'])
            kw_service.setdefault('comment', fwr.get('name', ''))
        return kw_service

    def _delete_fwr_service(self, context, vdom, **fwr):
        LOG.debug("# _delete_fwr_service() fwr=%(fwr)s", {'fwr': fwr})
        if fwr['protocol'] in ['any', None] and \
            not fwr['destination_port'] and not fwr['source_port']:
            return None
        else:
            return utils.delete_fwservice(
                self, context, vdom=vdom, name=fwr['id'])

    def _get_fwr_action(self, **fwr):
        if fwr.get('action', None) in ['allow']:
            action = 'accept'
        else:
            action = 'deny'
        return action

    def _get_fwp_profiles(self, action):
        profiles = {
            'av_profile': None,
            'webfilter_profile': None,
            'ips_sensor': None,
            'application_list': None,
            'ssl_ssh_profile': None
        }
        if action in ['allow', 'accept']:
            for key in profiles:
                profiles[key] = self._fortigate[key]
        return profiles

    def _get_fip_before_id(self, context, fwr_id):
        fwp_assed = fortinet_db.Fortinet_FW_Rule_Association.query_one(
                context, type=constants.TYPE_EXT, fwr_id=fwr_id)
        if not fwp_assed:
            fwp_assed = fortinet_db.Fortinet_FW_Rule_Association.query_one(
                context, type=constants.TYPE_INT, fwr_id=fwr_id)
        fwp = fortinet_db.query_record(context,
                                       fortinet_db.Fortinet_Firewall_Policy,
                                       id=fwp_assed.fortinet_pid)
        return getattr(fwp, 'edit_id', None)

    def _get_fips_in_fw(self, context, tenant_id, fw_net):
        fw_fips = []
        if not fw_net:
            return fw_fips
        namespace = fortinet_db.Fortinet_ML2_Namespace.query_one(
            context, tenant_id=tenant_id)
        if not namespace:
            return fw_fips
        db_fips = fortinet_db.query_records(
            context, l3_db.FloatingIP, tenant_id=tenant_id,
            status=l3_consts.FLOATINGIP_STATUS_ACTIVE)
        for fip in db_fips:
            if getattr(fip, 'fixed_ip_address', None) and \
                    IPAddress(fip.fixed_ip_address) in IPNetwork(fw_net):
                fw_fips.append((fip.id, fip.floating_ip_address))
        return fw_fips



# Copyright 2015 Fortinet, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Implentation of Fortinet ML2 Mechanism driver for ML2 Plugin."""

import sys

import netaddr
from oslo_log import log as logging

from neutron.common import constants as l3_constants
from neutron.db import api as db_api
from neutron.db import external_net_db as ext_db
from neutron.db import models_v2
from neutron.extensions import portbindings

from neutron.plugins.common import constants as p_const
from neutron.plugins.ml2.common import exceptions as ml2_exc
from neutron.plugins.ml2 import driver_api

from networking_fortinet._i18n import _LE, _LI
from networking_fortinet.common import config
from networking_fortinet.common import constants as const
from networking_fortinet.common import resources as resources
from networking_fortinet.common import utils as utils
from networking_fortinet.db import models as fortinet_db
from networking_fortinet.tasks import constants as t_consts
from networking_fortinet.tasks import tasks


LOG = logging.getLogger(__name__)


class FortinetMechanismDriver(driver_api.MechanismDriver):
    """ML2 Mechanism driver for Fortinet devices."""

    def __init__(self):
        super(FortinetMechanismDriver, self).__init__()

        self._driver = None
        self._fortigate = None
        self.task_manager = tasks.TaskManager()
        self.task_manager.start()

    def check_segment_for_agent(self, segment, agent):
            mappings = agent['configurations'].get('bridge_mappings', {})
            tunnel_types = agent['configurations'].get('tunnel_types', [])
            LOG.debug("Checking segment: %(segment)s "
                      "for mappings: %(mappings)s "
                      "with tunnel_types: %(tunnel_types)s",
                      {'segment': segment, 'mappings': mappings,
                       'tunnel_types': tunnel_types})
            network_type = segment[driver_api.NETWORK_TYPE]
            if network_type == 'local':
                return True
            elif network_type in tunnel_types:
                return True
            elif network_type in ['flat', 'vlan']:
                return segment[driver_api.PHYSICAL_NETWORK] in mappings
            else:
                return False

    def initialize(self):
        """Initilize of variables needed by this class."""
        self.Fortinet_init()

    def Fortinet_init(self):
        """Fortinet specific initialization for this class."""
        LOG.debug("FortinetMechanismDriver_init")
        self._fortigate = config.fgt_info
        self._driver = config.get_apiclient()

        for key in const.FORTINET_PARAMS:
            self.sync_conf_to_db(key)

        session = db_api.get_session()
        try:
            utils.add_vdom(self, session, vdom=const.EXT_VDOM,
                           tenant_id=const.FAKE_TENANT_ID)
            utils.set_vlanintf(self, session, vdom=const.EXT_VDOM,
                               name=self._fortigate['ext_interface'])
        except Exception as e:
            utils._rollback_on_err(self, session, e)
            raise ml2_exc.MechanismDriverError(
                method=sys._getframe().f_code.co_name)
        utils.update_status(self, session, t_consts.TaskStatus.COMPLETED)

    def sync_conf_to_db(self, param):
        cls = getattr(fortinet_db, const.FORTINET_PARAMS[param]['cls'])
        conf_list = self.get_range(param)
        session = db_api.get_session()
        records = fortinet_db.query_records(session, cls)
        for record in records:
            kwargs = {}
            for key in const.FORTINET_PARAMS[param]['keys']:
                _element = const.FORTINET_PARAMS[param]['type'](record[key])
                if _element not in conf_list and not record.allocated:
                    kwargs.setdefault(key, record[key])
                    fortinet_db.delete_record(session, cls, **kwargs)
        try:
            for i in range(0, len(conf_list),
                           len(const.FORTINET_PARAMS[param]['keys'])):
                kwargs = {}
                for key in const.FORTINET_PARAMS[param]['keys']:
                    kwargs.setdefault(key, str(conf_list[i]))
                    i += 1
                cls.init_records(session, **kwargs)
        except IndexError:
            LOG.error(_LE("The number of the configure range is not even,"
                        "the last one of %(param)s can not be used"),
                      {'param': param})
            raise IndexError

    def get_range(self, param):
        _type = const.FORTINET_PARAMS[param]['type']
        if const.FORTINET_PARAMS[param]['format']:
            min, max = self._fortigate[param].split(const.FIELD_DELIMITER)
            if _type(min) > _type(max):
                min, max = max, min
            if _type == int:
                min, max = _type(min), _type(max) + 1
            result = const.FORTINET_PARAMS[param]['range'](min, max)
        else:
            result = const.FORTINET_PARAMS[param]['range'](
                                _type(self._fortigate[param]),
                                const.FORTINET_PARAMS[param]['netmask'])

        return result if isinstance(result, list) else list(result)

    def create_network_precommit(self, mech_context):
        """Create Network in the mechanism specific database table."""
        LOG.debug("create_network_precommit: called")
        pass

    def create_network_postcommit(self, mech_context):
        """Create Network as a portprofile on the fortigate."""
        LOG.debug("create_network_postcommit: called")
        network = mech_context.current
        if network["router:external"]:
            # TODO(samsu)
            return
        # use network_id to get the network attributes
        # ONLY depend on our db for getting back network attributes
        # this is so we can replay postcommit from db
        network_name = network['name']
        tenant_id = network['tenant_id']
        segment = mech_context.network_segments[0]
        # currently supports only one segment per network
        if segment['network_type'] != 'vlan':
            raise Exception(_("Fortinet Mechanism: failed to create network,"
                              "only network type vlan is supported"))

        vlanid = segment['segmentation_id']
        context = mech_context._plugin_context
        try:
            namespace = utils.add_vdom(self, context, tenant_id=tenant_id)
            if not namespace:
                raise
            # TODO(samsu): type driver support vlan only,
            # need to check later
            inf_name = const.PREFIX['inf'] + str(vlanid)
            utils.add_vlanintf(self, context,
                               name=inf_name,
                               vdom=namespace.vdom,
                               vlanid=vlanid,
                               interface=self._fortigate['int_interface'],
                               alias=network_name)
        except Exception as e:
            utils._rollback_on_err(self, context, e)
            raise ml2_exc.MechanismDriverError(
                method=sys._getframe().f_code.co_name)
        utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)

    def delete_network_precommit(self, mech_context):
        """Delete Network from the plugin specific database table."""
        LOG.debug("delete_network_precommit: called")
        network = mech_context.current
        network_id = network['id']
        context = mech_context._plugin_context
        if fortinet_db.query_record(context, ext_db.ExternalNetwork,
                                    network_id=network_id):
            # return when the network is external network
            # TODO(samsu): may check external network
            # before delete namespace
            return
        tenant_id = network['tenant_id']
        namespace = fortinet_db.query_record(context,
                                    fortinet_db.Fortinet_ML2_Namespace,
                                    tenant_id=tenant_id)
        if not namespace:
            return
        # TODO(samsu): type driver support vlan only,
        # need to check later
        vlanid = network['provider:segmentation_id']
        inf_name = const.PREFIX['inf'] + str(vlanid)
        try:
            utils.delete_vlanintf(self, context, name=inf_name,
                     vdom=namespace.vdom)
        except Exception as e:
            resources.Exinfo(e)
            raise ml2_exc.MechanismDriverError(
                method=sys._getframe().f_code.co_name)

    def delete_network_postcommit(self, mech_context):
        """Delete network which translates to remove vlan interface
        and related vdom from the fortigate.
        """
        LOG.debug("delete_network_postcommit: called")
        network = mech_context.current
        context = mech_context._plugin_context
        tenant_id = network['tenant_id']
        with context.session.begin(subtransactions=True):
            try:
                utils.delete_vdom(self, context, tenant_id=tenant_id)
                LOG.info(_LI("delete network postcommit: tenant= %(tenant_id)s"
                           " network= %(network)s"),
                         {'tenant_id': tenant_id, 'network': network})
            except Exception as e:
                resources.Exinfo(e)
                raise ml2_exc.MechanismDriverError(
                    method=sys._getframe().f_code.co_name)

    def update_network_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        cur_network = mech_context.current
        org_network = mech_context.original
        if cur_network["router:external"] != org_network["router:external"]:
            LOG.info(_LI("update_network_precommit failed: the external "
                         "attribute cannot be updated, instead of updating the"
                         " attribute, the external network only support "
                         "creation from scratch on the admin view. \nThe "
                         "org_network: %(org)s, \nthe cur_network %(cur)s"),
                     {'org': org_network, 'cur': cur_network})
            raise NotImplementedError("The external attribute cannot be "
                                      "updated")
        pass

    def update_network_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        pass

    def create_subnet_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("create_subnetwork_precommit: called")

    def create_subnet_postcommit(self, mech_context, update=False):
        if not update:
            LOG.debug("create_subnetwork_postcommit: called")
        gateway = mech_context.current['gateway_ip']
        network_id = mech_context.current['network_id']
        subnet_id = mech_context.current['id']
        tenant_id = mech_context.current['tenant_id']
        context = mech_context._plugin_context
        dns_nameservers = mech_context.current.setdefault(
            'dns_nameservers', [])
        if update:
            router_func = utils.set_routerstatic
            dhcp_func = utils.set_dhcpserver
        else:
            router_func = utils.add_routerstatic
            dhcp_func = utils.add_dhcpserver
        try:
            if fortinet_db.query_record(context, ext_db.ExternalNetwork,
                                        network_id=network_id):

                router_func(self, context,
                            subnet_id=subnet_id,
                            vdom=const.EXT_VDOM,
                            dst=const.EXT_DEF_DST,
                            device=self._fortigate['ext_interface'],
                            gateway=gateway)
            else:
                namespace = fortinet_db.query_record(context,
                                        fortinet_db.Fortinet_ML2_Namespace,
                                        tenant_id=tenant_id)
                interface = utils.get_intf(context,
                                           mech_context.current['network_id'])
                netmask = str(netaddr.
                            IPNetwork(mech_context.current['cidr']).netmask)
                start_ip = mech_context.current['allocation_pools'][0]['start']
                end_ip = mech_context.current['allocation_pools'][0]['end']
                dhcp_func(self, context,
                          subnet_id=subnet_id,
                          vdom=namespace.vdom,
                          interface=interface,
                          dns_nameservers=dns_nameservers,
                          gateway=gateway,
                          netmask=netmask,
                          start_ip=start_ip,
                          end_ip=end_ip)

                # TODO(samsu): need to add rollback for the update and set
                cls = fortinet_db.Fortinet_Interface
                record = fortinet_db.query_record(context, cls,
                                                  name=interface,
                                                  vdom=namespace.vdom)
                if gateway:
                    cls.update_record(context, record,
                                      ip="%s %s" % (gateway, netmask))
                    utils.op(self, context, resources.VlanInterface.set,
                             name=interface,
                             vdom=namespace.vdom,
                             ip="%s %s" % (gateway, netmask))
        except Exception as e:
            utils._rollback_on_err(self, context, e)
            raise ml2_exc.MechanismDriverError(
                method=sys._getframe().f_code.co_name)
        if not update:
            utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)

    def delete_subnet_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("delete_subnetwork_precommit: called")

    def delete_subnet_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("delete_subnet_postcommit: called")
        context = mech_context._plugin_context
        subnet_id = mech_context.current['id']
        try:
            utils.delete_routerstatic(self, context, subnet_id=subnet_id)
            utils.delete_dhcpserver(self, context, subnet_id=subnet_id)
        except Exception as e:
            resources.Exinfo(e)
            raise ml2_exc.MechanismDriverError(
                method=sys._getframe().f_code.co_name)

    def update_subnet_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("update_subnet_precommit: called")

    def update_subnet_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("update_subnet_postcommit: called")
        self.create_subnet_postcommit(mech_context, update=True)

    def create_port_precommit(self, mech_context):
        """Create logical port on the fortigate (db update)."""
        LOG.debug("create_port_precommit: called")
        port = mech_context.current
        LOG.debug("create_port_precommit mech_context = %s", mech_context)
        context = mech_context._plugin_context
        namespace = fortinet_db.query_record(context,
                            fortinet_db.Fortinet_ML2_Namespace,
                            tenant_id=port['tenant_id'])
        port_id = port['id']
        subnet_id = port['fixed_ips'][0]['subnet_id']
        ip_address = port['fixed_ips'][0]['ip_address']
        mac = port['mac_address']
        db_subnetv2 = fortinet_db.query_record(context, models_v2.Subnet,
                                             id=subnet_id)
        if port['device_owner'] in ['network:router_gateway']:
            if fortinet_db.query_record(context, ext_db.ExternalNetwork,
                                        network_id=port['network_id']):
                utils.set_ext_gw(self, context, port)
        elif port['device_owner'] in ['compute:nova', 'compute:None', '']:
            # add dhcp related functions
            # '': create port before associate the port with a vm
            utils.add_reservedip(self, context,
                                 port_id=port_id,
                                 subnet_id=subnet_id,
                                 mac=mac,
                                 ip=ip_address,
                                 vdom=namespace.vdom)

        elif port['device_owner'] in ['network:router_interface']:
            if db_subnetv2.cidr:
                cidr = netaddr.IPNetwork(db_subnetv2.cidr)
                subnet = ' '.join([str(cidr.network), str(cidr.netmask)])
                utils.add_fwaddress(self, context,
                                   vdom=namespace.vdom,
                                   name=str(cidr.network),
                                   subnet=subnet)
                addrgrp_name = const.PREFIX['addrgrp'] + namespace.vdom
                utils.add_addrgrp(self, context,
                                  name=addrgrp_name,
                                  vdom=namespace.vdom,
                                  members=[str(cidr.network)])

                utils.add_fwpolicy(self, context,
                                   vdom=namespace.vdom,
                                   srcintf='any',
                                   srcaddr=addrgrp_name,
                                   dstintf='any',
                                   dstaddr=addrgrp_name,
                                   nat='disable')
        return

    def create_port_postcommit(self, mech_context):
        """Associate the assigned MAC address to the portprofile."""
        LOG.debug("create_port_postcommit: called")
        context = mech_context._plugin_context
        utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)

    def delete_port_postcommit(self, mech_context):
        LOG.debug("delete_port_postcommit: called")
        port = mech_context.current
        context = mech_context._plugin_context
        try:
            port_id = port['id']
            subnet_id = port['fixed_ips'][0]['subnet_id']
            db_subnet = fortinet_db.query_record(context,
                                             fortinet_db.Fortinet_ML2_Subnet,
                                             subnet_id=subnet_id)
            db_subnetv2 = fortinet_db.query_record(context, models_v2.Subnet,
                                                   id=subnet_id)
            if port['device_owner'] in ['network:router_gateway']:
                if fortinet_db.query_record(context, ext_db.ExternalNetwork,
                                            network_id=port['network_id']):
                    #delete ippool and its related firewall policy
                    utils.clr_ext_gw(self, context, port)

            elif port['device_owner'] in ['compute:nova', 'compute:None', '']:
                # delete dhcp related functions
                utils.delete_reservedip(self, context, port_id=port_id)

            elif port['device_owner'] in ['network:router_interface']:
                # add firewall address and address group
                name = const.PREFIX['addrgrp'] + db_subnet.vdom
                member = str(netaddr.IPNetwork(db_subnetv2.cidr).network)
                utils.delete_fwpolicy(self, context,
                                      vdom=db_subnet.vdom,
                                      srcintf='any',
                                      srcaddr=name,
                                      dstintf='any',
                                      dstaddr=name,
                                      nat='disable')
                utils.delete_addrgrp(self, context,
                                     name=name,
                                     vdom=db_subnet.vdom,
                                     members=member.split(' '))
                utils.delete_fwaddress(self, context,
                                       vdom=db_subnet.vdom,
                                       name=member)
        except Exception as e:
            resources.Exinfo(e)
            raise ml2_exc.MechanismDriverError(
                method=sys._getframe().f_code.co_name)

    def update_port_precommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("update_port_precommit: called")

    def update_port_postcommit(self, mech_context):
        """Noop now, it is left here for future."""
        LOG.debug("update_port_postcommit: called")

    def bind_port(self, context):
        """Marks ports as bound.

        Binds external ports and ports.
        Fabric configuration will occur on the subsequent port update.
        Currently only vlan segments are supported.
        """
        LOG.debug("bind_port() called")
        if (context.current['device_owner'] ==
            l3_constants.DEVICE_OWNER_ROUTER_INTF):
            # check controller to see if the port exists
            # so this driver can be run in parallel with others that add
            # support for external port bindings
            for segment in context.network.network_segments:
                if segment[driver_api.NETWORK_TYPE] == p_const.TYPE_VLAN:
                    context.set_binding(
                        segment[driver_api.ID], portbindings.VIF_TYPE_OVS,
                        {portbindings.CAP_PORT_FILTER: False,
                         portbindings.OVS_HYBRID_PLUG: False})
                    return

        for segment in context.network.network_segments:
            if segment[driver_api.NETWORK_TYPE] == p_const.TYPE_VLAN:
                context.set_binding(
                    segment[driver_api.ID], portbindings.VIF_TYPE_OVS,
                    {portbindings.CAP_PORT_FILTER: False,
                    portbindings.OVS_HYBRID_PLUG: False})



# Copyright 2015 Fortinet, Inc.
#
# All Rights Reserved
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

try:
    import httplib
except ImportError:
    import http.client as httplib

import eventlet
from oslo_log import log as logging
from oslo_serialization import jsonutils

from networking_fortinet._i18n import _LI, _LW
from networking_fortinet.api_client import request
from networking_fortinet.api_client import templates

LOG = logging.getLogger(__name__)
USER_AGENT = "Neutron eventlet client/2.0"


class EventletApiRequest(request.ApiRequest):
    '''Eventlet-based ApiRequest class.

    This class will form the basis for eventlet-based ApiRequest classes
    '''

    # Maximum number of green threads present in the system at one time.
    API_REQUEST_POOL_SIZE = request.DEFAULT_API_REQUEST_POOL_SIZE

    # Pool of green threads. One green thread is allocated per incoming
    # request. Incoming requests will block when the pool is empty.
    API_REQUEST_POOL = eventlet.GreenPool(API_REQUEST_POOL_SIZE)

    # A unique id is assigned to each incoming request. When the current
    # request id reaches MAXIMUM_REQUEST_ID it wraps around back to 0.
    MAXIMUM_REQUEST_ID = request.DEFAULT_MAXIMUM_REQUEST_ID

    # The request id for the next incoming request.
    CURRENT_REQUEST_ID = 0

    def __init__(self, client_obj, url, method="GET", body=None,
                 headers=None,
                 retries=request.DEFAULT_RETRIES,
                 auto_login=True,
                 redirects=request.DEFAULT_REDIRECTS,
                 http_timeout=request.DEFAULT_HTTP_TIMEOUT, client_conn=None):
        '''Constructor.'''
        self._api_client = client_obj
        self._url = url
        self._method = method
        self._body = body if body else None
        self._headers = headers or {}
        self._request_timeout = http_timeout * retries
        self._retries = retries
        self._auto_login = auto_login
        self._redirects = redirects
        self._http_timeout = http_timeout
        self._client_conn = client_conn
        self._abort = False

        self._request_error = None

        if "User-Agent" not in self._headers:
            self._headers["User-Agent"] = USER_AGENT

        self._green_thread = None
        # Retrieve and store this instance's unique request id.
        self._request_id = EventletApiRequest.CURRENT_REQUEST_ID
        # Update the class variable that tracks request id.
        # Request IDs wrap around at MAXIMUM_REQUEST_ID
        next_request_id = self._request_id + 1
        next_request_id %= self.MAXIMUM_REQUEST_ID
        EventletApiRequest.CURRENT_REQUEST_ID = next_request_id

    @classmethod
    def _spawn(cls, func, *args, **kwargs):
        '''Allocate a green thread from the class pool.'''
        return cls.API_REQUEST_POOL.spawn(func, *args, **kwargs)

    def spawn(self, func, *args, **kwargs):
        '''Spawn a new green thread with the supplied function and args.'''
        return self.__class__._spawn(func, *args, **kwargs)

    def join(self):
        '''Wait for instance green thread to complete.'''
        if self._green_thread is not None:
            return self._green_thread.wait()
        return Exception(_('Joining an invalid green thread'))

    def start(self):
        '''Start request processing.'''
        self._green_thread = self.spawn(self._run)

    def _run(self):
        '''Method executed within green thread.'''
        if self._request_timeout:
            # No timeout exception escapes the with block.
            with eventlet.timeout.Timeout(self._request_timeout, False):
                return self._handle_request()

            LOG.info(_LI('[%d] Request timeout.'), self._rid())
            self._request_error = Exception(_('Request timeout'))
            return None
        else:
            return self._handle_request()

    def _handle_request(self):
        '''First level request handling.'''
        attempt = 0
        timeout = 0
        response = None
        while response is None and attempt <= self._retries:
            eventlet.greenthread.sleep(timeout)
            attempt += 1

            req = self._issue_request()
            # automatically raises any exceptions returned.
            if isinstance(req, httplib.HTTPResponse):
                timeout = 0
                if attempt <= self._retries and not self._abort:
                    if req.status in (401, 403):
                        continue
                    elif req.status == 503:
                        timeout = 0.5
                        continue
                    # else fall through to return the error code

                LOG.debug("[%(rid)d] Completed request '%(method)s %(url)s'"
                          ": %(status)s",
                          {'rid': self._rid(), 'method': self._method,
                           'url': self._url, 'status': req.status})
                self._request_error = None
                response = req
            else:
                LOG.info(_LI('[%(rid)d] Error while handling request: '
                             '%(req)s'),
                         {'rid': self._rid(), 'req': req})
                self._request_error = req
                response = None
        return response


class LoginRequestEventlet(EventletApiRequest):
    '''Process a login request.'''

    def __init__(self, client_obj, user, password, client_conn=None,
                 headers=None):
        if headers is None:
            headers = {}
        headers.update({"Content-Type": "application/x-www-form-urlencoded"})
        message = client_obj._render(templates.LOGIN,
                                     username=user,
                                     secretkey=password)
        body = message['body']
        super(LoginRequestEventlet, self).__init__(
            client_obj, message['path'], message['method'], body, headers,
            auto_login=True, client_conn=client_conn)

    def session_cookie(self):
        if self.successful():
            return self.value.getheader("Set-Cookie")
        return None


class GetApiProvidersRequestEventlet(EventletApiRequest):
    '''Get a list of API providers.'''

    def __init__(self, client_obj):
        url = "/"
        super(GetApiProvidersRequestEventlet, self).__init__(
            client_obj, url, "GET", auto_login=True)

    def api_providers(self):
        """Parse api_providers from response.

        Returns: api_providers in [(host, port, is_ssl), ...] format
        """
        def _provider_from_listen_addr(addr):
            # (pssl|ptcp):<ip>:<port> => (host, port, is_ssl)
            parts = addr.split(':')
            return (parts[1], int(parts[2]), parts[0] == 'pssl')

        try:
            if self.successful():
                ret = []
                body = jsonutils.loads(self.value.body)
                for node in body.get('results', []):
                    for role in node.get('roles', []):
                        if role.get('role') == 'api_provider':
                            addr = role.get('listen_addr')
                            if addr:
                                ret.append(_provider_from_listen_addr(addr))
                return ret
        except Exception as e:
            LOG.warn(_LW("[%(rid)d] Failed to parse API provider: %(e)s"),
                     {'rid': self._rid(), 'e': e})
            # intentionally fall through
        return None


class GenericRequestEventlet(EventletApiRequest):
    '''Handle a generic request.'''

    def __init__(self, client_obj, method, url, body, content_type,
                 auto_login=False,
                 http_timeout=request.DEFAULT_HTTP_TIMEOUT,
                 retries=request.DEFAULT_RETRIES,
                 redirects=request.DEFAULT_REDIRECTS):
        headers = {"Content-Type": content_type}
        super(GenericRequestEventlet, self).__init__(
            client_obj, url, method, body, headers,
            retries=retries,
            auto_login=auto_login, redirects=redirects,
            http_timeout=http_timeout)

    def session_cookie(self):
        if self.successful():
            return self.value.getheader("Set-Cookie")
        return None


request.ApiRequest.register(EventletApiRequest)



# Copyright 2015 Fortinet, Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


"""Fortinet specific database schema/model."""

import copy
from neutron.db import model_base
from neutron.db import models_v2
from oslo_db import exception as os_db_exception
from oslo_log import log as logging
import six
import sqlalchemy as sa
from sqlalchemy.inspection import inspect
from sqlalchemy import orm

from networking_fortinet.common import constants as const


LOG = logging.getLogger(__name__)


OPS = ["ADD", "UPDATE", "DELETE", "QUERY"]


def add_record(context, cls, **kwargs):
    try:
        return cls.add_record(context, **kwargs)
    except os_db_exception.DBDuplicateEntry:
        pass
    return {}


def delete_record(context, cls, **kwargs):
    return cls.delete_record(context, kwargs)


def update_record(context, record, **kwargs):
    session = get_session(context)
    try:
        for key, value in six.iteritems(kwargs):
            if getattr(record, key, None) != value:
                setattr(record, key, value)
        with session.begin(subtransactions=True):
            session.add(record)
            return record
    except Exception as e:
        raise os_db_exception.DBError(e)


def query_record(context, cls, **kwargs):
    return db_query(cls, context, **kwargs).first()


def query_records(context, cls, **kwargs):
    return db_query(cls, context, **kwargs).all()


def query_count(context, cls, **kwargs):
    return db_query(cls, context, **kwargs).count()


def get_session(context):
    return context.session if getattr(context, 'session', None) else context


def primary_keys(cls):
    """
    :param cls: Object or instance derived from the class ModelBase
                in the library sqlalchemy
    :return: list of primary keys
    """
    if not isinstance(cls, type):
        cls = cls.__class__
    return [key.name for key in inspect(cls).primary_key]


def db_query(cls, context, lockmode=False, **kwargs):
    """
    :param cls:
    :param context:
    :param kwargs:
    :return:
    """
    session = get_session(context)
    query = session.query(cls)
    if lockmode:
        query = query.with_lockmode('update')
    for key, value in six.iteritems(kwargs):
        kw = {key: value}
        query = query.filter_by(**kw)
    return query


class DBbase(object):
    @classmethod
    def init_records(cls, context, **kwargs):
        """Add records to be allocated into the table."""
        session = get_session(context)
        with session.begin(subtransactions=True):
            record = cls.query_one(context, **kwargs)
            if not record:
                record = cls()
                for key, value in six.iteritems(kwargs):
                    if hasattr(record, key):
                        setattr(record, key, value)
                session.add(record)

    @classmethod
    def add_record(cls, context, **kwargs):
        """Add vlanid to be allocated into the table."""
        session = get_session(context)
        _kwargs = copy.copy(kwargs)
        for key in _kwargs:
            if not hasattr(cls, key):
                kwargs.pop(key, None)
        with session.begin(subtransactions=True):
            record = cls.query_one(context, **kwargs)
            if not record:
                record = cls()
                for key, value in six.iteritems(kwargs):
                    setattr(record, key, value)
                session.add(record)
                rollback = cls._prepare_rollback(context,
                                                 cls.delete_record,
                                                 **kwargs)
            else:
                rollback = {}
                #raise os_db_exception.DBDuplicateEntry
        ## TODO(samsu): kwargs would be better if only include class cls
        ## related primary keys
        return {'result': record, 'rollback': rollback}

    @staticmethod
    def update_record(context, record, **kwargs):
        """Add vlanid to be allocated into the table."""
        return update_record(context, record, **kwargs)

    @classmethod
    def delete_record(cls, context, kwargs):
        """
        Notes: kwargs is a dictionary as a variable, no wrapped as
        **kwargs because of the requirements of rollback process.
        Delete the record with the value of kwargs from the database,
        kwargs is a dictionary variable, here should not pass into a
        variable like **kwargs
        """
        session = get_session(context)
        with session.begin(subtransactions=True):
            record = cls.query_one(context, lockmode=True, **kwargs)
            if record:
                session.delete(record)
        return record

    @classmethod
    def query_one(cls, context, lockmode=False, **kwargs):
        """Get a filtered vlink_vlan_allocation record."""
        query = db_query(cls, context, lockmode=lockmode, **kwargs)
        return query.first()

    @classmethod
    def query_all(cls, context, **kwargs):
        """Get a filtered vlink_vlan_allocation record."""
        query = db_query(cls, context, **kwargs)
        return query.all()

    @classmethod
    def query_count(cls, context, **kwargs):
        """Get a filtered vlink_vlan_allocation record."""
        query = db_query(cls, context, **kwargs)
        return query.count()

    @classmethod
    def query_like(cls, context, **kwargs):
        """Get a filtered vlink_vlan_allocation record."""
        session = get_session(context)
        query = session.query(cls)
        for key, value in six.iteritems(kwargs):
            if getattr(cls, key, None):
                query = query.filter(getattr(cls, key, None).like(value))
        return query.all()

    @staticmethod
    def _prepare_rollback(context, func, **kwargs):
        if not func:
            raise ValueError
        rollback = {
            'func': func,
            'params': (context, kwargs)
        }
        return rollback


class Fortinet_ML2_Namespace(model_base.BASEV2, DBbase):
    """Schema for Fortinet network."""
    id = sa.Column(sa.Integer, primary_key=True, autoincrement=True)
    tenant_id = sa.Column(sa.String(36), primary_key=True)
    # For the name of vdom has the following restrictions:
    # only letters, numbers, "-" and "_" are allowed
    # no more than 11 characters are allowed
    # no spaces are allowed
    vdom = sa.Column(sa.String(11))

    @classmethod
    def add_record(cls, context, **kwargs):
        res = super(Fortinet_ML2_Namespace, cls).add_record(context, **kwargs)
        if res.get('rollback'):
            res['result']._allocate_vdom(context, res['result'])
        return res

    def _allocate_vdom(self, context, record):
        if not getattr(record, 'vdom'):
            vdom = const.PREFIX['vdom'] + str(record.id)
            self.update_record(context, record, vdom=vdom)
        return record.vdom


class Fortinet_ML2_Subnet(model_base.BASEV2, DBbase):
    """Schema to map subnet to Fortinet dhcp interface."""
    subnet_id = sa.Column(sa.String(36), primary_key=True)
    vdom = sa.Column(sa.String(11))
    interface = sa.Column(sa.String(11), default=None)
    gateway = sa.Column(sa.String(32), default=None)
    netmask = sa.Column(sa.String(32), default=None)
    start_ip = sa.Column(sa.String(32), default=None)
    end_ip = sa.Column(sa.String(32), default=None)
    edit_id = sa.Column(sa.Integer)


class Fortinet_ML2_ReservedIP(model_base.BASEV2, DBbase):
    """Schema for Fortinet dhcp server reserved ip."""
    port_id = sa.Column(sa.String(36), primary_key=True)
    subnet_id = sa.Column(sa.String(36))
    mac = sa.Column(sa.String(32))
    ip = sa.Column(sa.String(32))
    vdom = sa.Column(sa.String(11))
    edit_id = sa.Column(sa.Integer)

    @classmethod
    def add_record(cls, context, **kwargs):
        res = super(Fortinet_ML2_ReservedIP, cls).add_record(context, **kwargs)
        if res.get('rollback'):
            cls._allocate_edit_id(context, res['result'])
        return res

    @classmethod
    def _allocate_edit_id(cls, context, record):
        if not getattr(record, 'edit_id'):
            last_record = (db_query(cls, context, subnet_id=record.subnet_id).
                           order_by(cls.edit_id.desc()).first())
            edit_id = last_record.edit_id + 1 if last_record.edit_id else 1
            record.update_record(context, record, edit_id=edit_id)
        return record.edit_id


class Fortinet_Static_Router(model_base.BASEV2, models_v2.HasId, DBbase):
    """Schema for Fortinet static router."""
    vdom = sa.Column(sa.String(11))
    subnet_id = sa.Column(sa.String(36))
    dst = sa.Column(sa.String(32))
    device = sa.Column(sa.String(32))
    gateway = sa.Column(sa.String(32))
    edit_id = sa.Column(sa.Integer)


class Fortinet_Vlink_Vlan_Allocation(model_base.BASEV2, models_v2.HasId,
                                     DBbase):
    """Schema for Fortinet vlink vlan interface."""
    vdom = sa.Column(sa.String(11))
    inf_name_int_vdom = sa.Column(sa.String(11))
    inf_name_ext_vdom = sa.Column(sa.String(11))
    vlanid = sa.Column(sa.Integer)
    allocated = sa.Column(sa.Boolean(), default=False, nullable=False)

    @staticmethod
    def reset():
        """
        set all value of keys in kwargs to the default value(None or False)
        """
        return {
            'vdom': None,
            'inf_name_int_vdom': None,
            'inf_name_ext_vdom': None,
            'allocated': False
        }

    @classmethod
    def add_record(cls, context, **kwargs):
        record = cls.query_one(context, **kwargs)
        if not record:
            kwargs.setdefault('allocated', True)
            if kwargs.get('vdom', None):
                kwargs.setdefault('inf_name_int_vdom',
                                  kwargs['vdom'] + const.POSTFIX['vint'])
                kwargs.setdefault('inf_name_ext_vdom',
                                  kwargs['vdom'] + const.POSTFIX['vext'])
            # TODO(samsu): if no vlanid needed, then it should able to add
            # a new record, consider to separate vlanid to a table.
            record = cls.query_one(context, lockmode=True, allocated=False)
            record.update_record(context, record, **kwargs)
            rollback = cls._prepare_rollback(context, cls.delete_record,
                                             **kwargs)
        else:
            rollback = {}
        ## need to check the attribute in the record whether updated
        ## # after update_record()
        return {'result': record, 'rollback': rollback}

    @classmethod
    def delete_record(cls, context, kwargs):
        """Delete vlanid to be allocated into the table."""
        session = get_session(context)
        with session.begin(subtransactions=True):
            record = cls.query_one(context, **kwargs)
            if record:
                record.update_record(context, record, **cls.reset())
        return record


class Fortinet_Vlink_IP_Allocation(model_base.BASEV2, DBbase):
    """Schema for Fortinet vlink vlan interface."""
    vlink_ip_subnet = sa.Column(sa.String(32), primary_key=True)
    vdom = sa.Column(sa.String(11))
    vlink_id = sa.Column(sa.String(36),
                         sa.ForeignKey("fortinet_vlink_vlan_allocations.id",
                                       ondelete="CASCADE"), nullable=True)
    allocated = sa.Column(sa.Boolean(), default=False, nullable=False)

    @staticmethod
    def reset():
        """
        set all value of keys in kwargs to the default value(None or False)
        """
        return {
            'vdom': None,
            'vlanid': None,
            'allocated': False
        }

    @classmethod
    def add_record(cls, context, **kwargs):
        session = get_session(context)
        with session.begin(subtransactions=True):
            record = cls.query_one(context, **kwargs)
            if not record:
                record = cls.query_one(context, lockmode=True, allocated=False)
                kwargs.setdefault('allocated', True)
                record.update_record(context, record, **kwargs)
                rollback = cls._prepare_rollback(context,
                                                 cls.delete_record,
                                                 **kwargs)
            else:
                rollback = {}
        ## need to check the attribute in the record whether updated
        ## # after update_record()
        return {'result': record, 'rollback': rollback}

    @classmethod
    def delete_record(cls, context, kwargs):
        """Delete vlanid to be allocated into the table."""
        session = get_session(context)
        with session.begin(subtransactions=True):
            record = cls.query_one(context, **kwargs)
            if record:
                record.update_record(context, record, **cls.reset())
        return record


class Fortinet_Vdom_Vlink(model_base.BASEV2, DBbase):
    """Schema for Fortinet vlink vlan interface."""
    name = sa.Column(sa.String(11), primary_key=True)
    vdom = sa.Column(sa.String(11), primary_key=True)

    @classmethod
    def add_record(cls, context, **kwargs):
        if kwargs.get('vdom', None):
            kwargs.setdefault('name', kwargs['vdom'] + const.POSTFIX['vdlink'])
        return super(Fortinet_Vdom_Vlink, cls).add_record(context, **kwargs)


class Fortinet_Firewall_Policy(model_base.BASEV2, models_v2.HasId, DBbase):
    """Schema for Fortinet firewall policy."""
    __tablename__ = 'fortinet_firewall_policies'
    vdom = sa.Column(sa.String(11))
    srcintf = sa.Column(sa.String(11), default="any")
    dstintf = sa.Column(sa.String(11), default="any")
    srcaddr = sa.Column(sa.String(40), default="all")
    dstaddr = sa.Column(sa.String(40), default="all")
    poolname = sa.Column(sa.String(32), default=None)
    nat = sa.Column(sa.String(7), default="disable")
    action = sa.Column(sa.String(11), default="accept")
    service = sa.Column(sa.String(36), default="ALL")
    match_vip = sa.Column(sa.String(7), default="disable")
    status = sa.Column(sa.String(7), default="enable")
    av_profile = sa.Column(sa.String(35), default=None)
    webfilter_profile = sa.Column(sa.String(35), default=None)
    ips_sensor = sa.Column(sa.String(35), default=None)
    application_list = sa.Column(sa.String(35), default=None)
    ssl_ssh_profile = sa.Column(sa.String(35), default=None)
    # comments(max 1023 in fortigate) to save firewall rule name (max 255)
    comments = sa.Column(sa.String(255), default=None)
    edit_id = sa.Column(sa.Integer)


class Fortinet_FloatingIP_Allocation(model_base.BASEV2, DBbase):
    """Schema for Fortinet vlink vlan interface.
    ip_subnet: it is a network with 30 bits network, there
    are two ips available, the smaller one will be allocated
    to the interface of the external network vdom and the
    bigger one will be allocated to the interface of related
    tenant network vdom.
    """
    ip_subnet = sa.Column(sa.String(32), primary_key=True)
    floating_ip_address = sa.Column(sa.String(36))
    vdom = sa.Column(sa.String(11))
    vip_name = sa.Column(sa.String(50))
    ## secondary_ip = sa.Column(sa.String(50), default=None)
    allocated = sa.Column(sa.Boolean(), default=False, nullable=False)
    ## TODO(samsu): delete the field bound

    @staticmethod
    def reset():
        """
        set all value of keys in kwargs to the default value(None or False)
        """
        return {
            'floating_ip_address': None,
            'vdom': None,
            'vip_name': None,
            'allocated': False
        }

    @classmethod
    def add_record(cls, context, **kwargs):
        session = get_session(context)
        with session.begin(subtransactions=True):
            record = cls.query_one(context, **kwargs)
            if not record:
                record = cls.query_one(context, lockmode=True, allocated=False)
                kwargs.setdefault('allocated', True)
                record.update_record(context, record, **kwargs)
                rollback = cls._prepare_rollback(context,
                                                 cls.delete_record,
                                                 **kwargs)
            else:
                rollback = {}
        ## need to check the attribute in the record whether updated
        ## # after update_record()
        return {'result': record, 'rollback': rollback}

    @classmethod
    def delete_record(cls, context, kwargs):
        """Delete vlanid to be allocated into the table."""
        session = get_session(context)
        with session.begin(subtransactions=True):
            record = cls.query_one(context, **kwargs)
            if record:
                record.update_record(context, record, **cls.reset())
        return record


class Fortinet_Firewall_VIP(model_base.BASEV2, DBbase):
    name = sa.Column(sa.String(36), primary_key=True)
    vdom = sa.Column(sa.String(11), primary_key=True)
    extip = sa.Column(sa.String(32))
    extintf = sa.Column(sa.String(32))
    mappedip = sa.Column(sa.String(32), default=None)


class Fortinet_Firewall_IPPool(model_base.BASEV2, DBbase):
    name = sa.Column(sa.String(36), primary_key=True)
    vdom = sa.Column(sa.String(11))
    startip = sa.Column(sa.String(32))
    endip = sa.Column(sa.String(32))
    type = sa.Column(sa.String(32), default="one-to-one")
    comments = sa.Column(sa.String(32), default=None)

    @classmethod
    def add_record(cls, context, **kwargs):
        kwargs.setdefault('endip', kwargs.get('startip', None))
        return super(Fortinet_Firewall_IPPool, cls).add_record(context,
                                                               **kwargs)


class Fortinet_Firewall_Address(model_base.BASEV2, DBbase):
    __tablename__ = 'fortinet_firewall_addresses'
    name = sa.Column(sa.String(40), primary_key=True)
    vdom = sa.Column(sa.String(11), primary_key=True)
    subnet = sa.Column(sa.String(32))
    associated_interface = sa.Column(sa.String(11), default=None)
    group = sa.Column(sa.String(32), default=None)


class Fortinet_Firewall_Service(model_base.BASEV2, DBbase):
    #__tablename__ = 'fortinet_firewall_services'
    # service name <- firewall_rule id
    name = sa.Column(sa.String(36), primary_key=True)
    vdom = sa.Column(sa.String(11), primary_key=True)


class Fortinet_Interface(model_base.BASEV2, DBbase):
    name = sa.Column(sa.String(36), primary_key=True)
    vdom = sa.Column(sa.String(11))
    vlanid = sa.Column(sa.Integer)
    interface = sa.Column(sa.String(11), default=None)
    type = sa.Column(sa.String(32), default=None)
    ip = sa.Column(sa.String(32), default=const.EXT_DEF_DST)
    secondary_ip = sa.Column(sa.String(11), default="enable")
    alias = sa.Column(sa.String(32), default=None)


class Fortinet_Interface_subip(model_base.BASEV2, DBbase):
    ip = sa.Column(sa.String(32), primary_key=True)
    name = sa.Column(sa.String(11), default=None)
    vdom = sa.Column(sa.String(11))


class Fortinet_FW_Rule_Association(model_base.BASEV2, DBbase):
    fwr_id = sa.Column(sa.String(36),
        sa.ForeignKey('firewall_rules.id', ondelete="CASCADE"),
        primary_key=True)
    fortinet_pid = sa.Column(sa.String(36),
        sa.ForeignKey('fortinet_firewall_policies.id', ondelete="CASCADE"),
        primary_key=True)
    type = sa.Column(sa.String(36), default=None)
    floatingip_id = sa.Column(sa.String(36),
        sa.ForeignKey('floatingips.id', ondelete="CASCADE"), nullable=True)
    fortinet_policy = orm.relationship(
        "Fortinet_Firewall_Policy", backref='fortinet_fw_rule_association')



# Copyright 2015 Fortinet Inc.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#


"""Implentation of FortiOS service Plugin."""

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from neutron.common import constants as l3_constants
from neutron.common import exceptions as n_exc
from neutron.db import l3_db
from neutron import manager
from neutron.plugins.common import constants as service_consts
from neutron.plugins.ml2 import db
from neutron.services.l3_router import l3_router_plugin as router

from networking_fortinet._i18n import _LE
from networking_fortinet.common import config
from networking_fortinet.common import constants as const
from networking_fortinet.common import resources
from networking_fortinet.common import utils
from networking_fortinet.db import models as fortinet_db
from networking_fortinet.tasks import constants as t_consts
from networking_fortinet.tasks import tasks

# TODO(samsu): the folowing two imports just for testing purpose
# TODO(samsu): need to be deleted later
from neutron.db import models_v2

DEVICE_OWNER_ROUTER_INTF = l3_constants.DEVICE_OWNER_ROUTER_INTF
DEVICE_OWNER_ROUTER_GW = l3_constants.DEVICE_OWNER_ROUTER_GW
DEVICE_OWNER_FLOATINGIP = l3_constants.DEVICE_OWNER_FLOATINGIP


LOG = logging.getLogger(__name__)


class FortinetL3ServicePlugin(router.L3RouterPlugin):
    """Fortinet L3 service Plugin."""

    supported_extension_aliases = ["router", "ext-gw-mode", "extraroute"]

    def __init__(self):
        """Initialize Fortinet L3 service Plugin."""
        super(FortinetL3ServicePlugin, self).__init__()
        self._fortigate = None
        self._driver = None
        self.task_manager = tasks.TaskManager()
        self.task_manager.start()
        self.Fortinet_init()

    def Fortinet_init(self):
        """Fortinet specific initialization for this class."""
        LOG.debug("FortinetL3ServicePlugin_init")
        self._fortigate = config.fgt_info
        self._driver = config.get_apiclient()
        self.enable_fwaas = 'fwaas_fortinet' in cfg.CONF.service_plugins

    def create_router(self, context, router):
        LOG.debug("create_router: router=%s" % (router))
        # Limit one router per tenant
        if not router.get('router', None):
            return
        tenant_id = router['router']['tenant_id']
        if fortinet_db.query_count(context, l3_db.Router,
                                   tenant_id=tenant_id):
            raise Exception(_("FortinetL3ServicePlugin:create_router "
                              "Only support one router per tenant"))
        with context.session.begin(subtransactions=True):
            try:
                namespace = utils.add_vdom(self, context, tenant_id=tenant_id)
                utils.add_vlink(self, context, namespace.vdom)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    LOG.error(_LE("Failed to create_router router=%(router)s"),
                              {"router": router})
                    utils._rollback_on_err(self, context, e)
        utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)
        return super(FortinetL3ServicePlugin, self).\
            create_router(context, router)

    def update_router(self, context, id, router):
        LOG.debug("update_router: id=%(id)s, router=%(router)s",
                  {'id': id, 'router': router})
        return (super(FortinetL3ServicePlugin, self).
                update_router(context, id, router))

    def delete_router(self, context, id):
        LOG.debug("delete_router: router id=%s" % (id))
        try:
            if self.enable_fwaas:
                fw_plugin = manager.NeutronManager.get_service_plugins().get(
                    service_consts.FIREWALL)
                fw_plugin.update_firewall_for_delete_router(context, id)
            with context.session.begin(subtransactions=True):
                router = fortinet_db.query_record(context, l3_db.Router, id=id)
                super(FortinetL3ServicePlugin, self).delete_router(context, id)
                if getattr(router, 'tenant_id', None):
                    utils.delete_vlink(self, context, router.tenant_id)
                    utils.delete_vdom(self, context,
                                      tenant_id=router.tenant_id)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                LOG.error(_LE("Failed to delete_router routerid=%(id)s"),
                          {"id": id})
                resources.Exinfo(e)

    def add_router_interface(self, context, router_id, interface_info):
        """creates vlnk on the fortinet device."""
        LOG.debug("FortinetL3ServicePlugin.add_router_interface: "
                  "router_id=%(router_id)s "
                  "interface_info=%(interface_info)r",
                  {'router_id': router_id, 'interface_info': interface_info})
        with context.session.begin(subtransactions=True):
            info = super(FortinetL3ServicePlugin, self).add_router_interface(
                context, router_id, interface_info)
            port = db.get_port(context.session, info['port_id'])
            port['admin_state_up'] = True
            port['port'] = port
            LOG.debug("FortinetL3ServicePlugin: "
                      "context=%(context)s"
                      "port=%(port)s "
                      "info=%(info)r",
                      {'context': context, 'port': port, 'info': info})
            interface_info = info
            subnet = self._core_plugin._get_subnet(context,
                                                   interface_info['subnet_id'])
            network_id = subnet['network_id']
            tenant_id = port['tenant_id']
            port_filters = {'network_id': [network_id],
                            'device_owner': [DEVICE_OWNER_ROUTER_INTF]}
            port_count = self._core_plugin.get_ports_count(context,
                                                           port_filters)
            # port count is checked against 2 since the current port is already
            # added to db
            if port_count == 2:
                # This subnet is already part of some router
                LOG.error(_LE("FortinetL3ServicePlugin: adding redundant "
                              "router interface is not supported"))
                raise Exception(_("FortinetL3ServicePlugin:adding redundant "
                                  "router interface is not supported"))
            try:
                db_namespace = fortinet_db.query_record(context,
                                        fortinet_db.Fortinet_ML2_Namespace,
                                        tenant_id=tenant_id)
                vlan_inf = utils.get_intf(context, network_id)
                int_intf, ext_intf = utils.get_vlink_intf(self, context,
                                               vdom=db_namespace.vdom)
                utils.add_fwpolicy(self, context,
                                   vdom=db_namespace.vdom,
                                   srcintf=vlan_inf,
                                   dstintf=int_intf,
                                   nat='enable')

            except Exception as e:
                LOG.error(_LE("Failed to create Fortinet resources to add "
                            "router interface. info=%(info)s, "
                            "router_id=%(router_id)s"),
                          {"info": info, "router_id": router_id})
                utils._rollback_on_err(self, context, e)
                with excutils.save_and_reraise_exception():
                    self.remove_router_interface(context, router_id,
                                                 interface_info)
        utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)
        return info

    def remove_router_interface(self, context, router_id, interface_info):
        """Deletes vlink, default router from Fortinet device."""
        LOG.debug("FortinetL3ServicePlugin.remove_router_interface called: "
                  "router_id=%(router_id)s "
                  "interface_info=%(interface_info)r",
                  {'router_id': router_id, 'interface_info': interface_info})

        with context.session.begin(subtransactions=True):
            info = (super(FortinetL3ServicePlugin, self).
                    remove_router_interface(context, router_id,
                                            interface_info))
            try:
                subnet = self._core_plugin._get_subnet(context,
                                                       info['subnet_id'])
                tenant_id = subnet['tenant_id']
                network_id = subnet['network_id']
                vlan_inf = utils.get_intf(context, network_id)
                db_namespace = fortinet_db.query_record(context,
                                        fortinet_db.Fortinet_ML2_Namespace,
                                        tenant_id=tenant_id)
                utils.delete_fwpolicy(self, context,
                                      vdom=db_namespace.vdom,
                                      srcintf=vlan_inf)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error(_LE("Fail remove of interface from Fortigate "
                                  "router interface. info=%(info)s, "
                                  "router_id=%(router_id)s"),
                             {"info": info, "router_id": router_id})
        return info

    def create_floatingip(self, context, floatingip):
        """Create floating IP.

        :param context: Neutron request context
        :param floatingip: data for the floating IP being created
        :returns: A floating IP object on success

        As the l3 router plugin asynchronously creates floating IPs
        leveraging the l3 agent, the initial status for the floating
        IP object will be DOWN.
        """
        LOG.debug("create_floatingip: floatingip=%s", floatingip)
        returned_obj = (super(FortinetL3ServicePlugin, self).
                        create_floatingip(context, floatingip))
        try:
            self._allocate_floatingip(context, returned_obj)
            if returned_obj.get('port_id', None):
                if not floatingip['floatingip'].get('fixed_ip_address', None):
                    floatingip['floatingip']['fixed_ip_address'] = (
                        returned_obj.get('fixed_ip_address'))
                self._associate_floatingip(context, returned_obj['id'],
                                           floatingip)
                self.update_floatingip_status(context, returned_obj,
                                    l3_constants.FLOATINGIP_STATUS_ACTIVE,
                                    id=returned_obj['id'])
            return returned_obj
        except Exception as e:
            with excutils.save_and_reraise_exception():
                resources.Exinfo(e)
                super(FortinetL3ServicePlugin, self).delete_floatingip(
                    context, returned_obj['id'])

    def delete_floatingip(self, context, id):
        LOG.debug("delete_floatingip called() id=%s", id)
        fip = fortinet_db.query_record(context, l3_db.FloatingIP, id=id)
        if fip and getattr(fip, 'fixed_port_id', None):
            self._disassociate_floatingip(context, id)
            super(FortinetL3ServicePlugin, self).disassociate_floatingips(
                context, fip['fixed_port_id'])
        self._release_floatingip(context, id)

    def update_floatingip_status(self, context, res, status, **kwargs):
        if res.get('status', None):
            res['status'] = status
        record = fortinet_db.query_record(context, l3_db.FloatingIP, **kwargs)
        fortinet_db.update_record(context, record, status=status)

    def update_floatingip(self, context, id, floatingip):
        if floatingip['floatingip']['port_id']:
            # floating ip associate with VM port.
            try:
                res = (super(FortinetL3ServicePlugin, self).
                       update_floatingip(context, id, floatingip))
                if not floatingip['floatingip'].get('fixed_ip_address', None):
                    floatingip['floatingip']['fixed_ip_address'] = (res.
                        get('fixed_ip_address'))
                self._associate_floatingip(context, id, floatingip)
                self.update_floatingip_status(context, res,
                                l3_constants.FLOATINGIP_STATUS_ACTIVE, id=id)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    resources.Exinfo(e)
                    super(FortinetL3ServicePlugin,
                          self).disassociate_floatingips(
                        context, floatingip['floatingip']['port_id'])
        else:
            # disassociate floating ip.
            self._disassociate_floatingip(context, id)
            res = (super(FortinetL3ServicePlugin, self).
                   update_floatingip(context, id, floatingip))
            self.update_floatingip_status(context, res,
                            l3_constants.FLOATINGIP_STATUS_DOWN, id=id)
        return res if res else None

    def _associate_floatingip(self, context, id, floatingip):
        try:
            l3db_fip = self._get_floatingip(context, id)
            db_namespace = fortinet_db.query_record(context,
                                    fortinet_db.Fortinet_ML2_Namespace,
                                    tenant_id=l3db_fip.tenant_id)

            db_fip = fortinet_db.query_record(context,
                            fortinet_db.Fortinet_FloatingIP_Allocation,
                            floating_ip_address=l3db_fip.floating_ip_address,
                            allocated=True)
            int_intf, ext_intf = utils.get_vlink_intf(self, context,
                                                      vdom=db_namespace.vdom)
            mappedip = utils.get_ipaddr(db_fip.ip_subnet, 0)
            fixed_ip_address = floatingip['floatingip']['fixed_ip_address']
            utils.add_vip(self, context,
                          vdom=db_namespace.vdom,
                          name=db_fip.floating_ip_address,
                          extip=mappedip,
                          extintf=int_intf,
                          mappedip=fixed_ip_address)

            db_ip = fortinet_db.query_record(context, models_v2.IPAllocation,
                                port_id=floatingip['floatingip']['port_id'])
            vlan_inf = utils.get_intf(context, db_ip.network_id)
            utils.add_fwpolicy(self, context,
                               vdom=db_namespace.vdom,
                               srcintf=int_intf,
                               dstintf=vlan_inf,
                               dstaddr=db_fip.floating_ip_address,
                               nat='enable')

            utils.add_fwaddress(self, context,
                                name=fixed_ip_address,
                                vdom=db_namespace.vdom,
                                subnet="%s 255.255.255.255" % fixed_ip_address,
                                associated_interface=vlan_inf)

            db_fwpolicy = utils.add_fwpolicy(self, context,
                               vdom=db_namespace.vdom,
                               srcintf=vlan_inf,
                               srcaddr=fixed_ip_address,
                               dstintf=int_intf,
                               poolname=mappedip)

            if self.enable_fwaas:
                fwrass = fortinet_db.Fortinet_FW_Rule_Association.query_one(
                    context, fwr_id=db_namespace.tenant_id)
                default_fwp = getattr(fwrass, 'fortinet_policy', None)
                if getattr(default_fwp, 'edit_id', None):
                    utils.head_firewall_policy(self, context,
                                               vdom=db_namespace.vdom,
                                               id=db_fwpolicy.edit_id,
                                               after=default_fwp.edit_id)
                    _headed = True
            if '_headed' not in locals():
                utils.head_firewall_policy(self, context,
                                           vdom=db_namespace.vdom,
                                           id=db_fwpolicy.edit_id)
        except Exception as e:
            with excutils.save_and_reraise_exception():
                utils._rollback_on_err(self, context, e)
        utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)

    def _disassociate_floatingip(self, context, id):
        l3db_fip = self._get_floatingip(context, id)
        db_namespace = fortinet_db.query_record(context,
                                    fortinet_db.Fortinet_ML2_Namespace,
                                    tenant_id=l3db_fip.tenant_id)
        db_fip = fortinet_db.query_record(context,
                            fortinet_db.Fortinet_FloatingIP_Allocation,
                            floating_ip_address=l3db_fip.floating_ip_address,
                            allocated=True)
        int_intf, ext_intf = utils.get_vlink_intf(self, context,
                                               vdom=db_namespace.vdom)
        db_ip = fortinet_db.query_record(context, models_v2.IPAllocation,
                                         port_id=l3db_fip.fixed_port_id)
        vlan_inf = utils.get_intf(context, db_ip.network_id)
        mappedip = utils.get_ipaddr(db_fip.ip_subnet, 0)
        utils.delete_fwpolicy(self, context,
                              vdom=db_namespace.vdom,
                              srcintf=vlan_inf,
                              srcaddr=l3db_fip.fixed_ip_address,
                              dstintf=int_intf,
                              poolname=mappedip)

        utils.delete_fwaddress(self, context,
                               name=l3db_fip.fixed_ip_address,
                               vdom=db_namespace.vdom)

        utils.delete_fwpolicy(self, context,
                              vdom=db_namespace.vdom,
                              dstaddr=db_fip.floating_ip_address)

        utils.delete_vip(self, context,
                         vdom=db_namespace.vdom,
                         name=db_fip.floating_ip_address)

    def disassociate_floatingips(self, context, port_id, do_notify=True):
        fip = fortinet_db.query_record(context, l3_db.FloatingIP,
                                       fixed_port_id=port_id)
        if fip and getattr(fip, 'id', None):
            self._disassociate_floatingip(context, fip.id)
        return super(FortinetL3ServicePlugin,
                     self).disassociate_floatingips(context,
                                                    port_id,
                                                    do_notify=do_notify)

    def _add_interface_by_subnet(self, context, router, subnet_id, owner):
        LOG.debug("_add_interface_by_subnet(): router=%(router)s, "
                  "subnet_id=%(subnet_id)s, owner=%(owner)s",
                  {'router': router, 'subnet_id': subnet_id, 'owner': owner})
        subnet = self._core_plugin._get_subnet(context, subnet_id)
        if not subnet['gateway_ip']:
            msg = _('Subnet for router interface must have a gateway IP')
            raise n_exc.BadRequest(resource='router', msg=msg)
        self._check_for_dup_router_subnet(context, router,
                                          subnet['network_id'],
                                          subnet_id,
                                          subnet['cidr'])
        fixed_ip = {'ip_address': subnet['gateway_ip'],
                    'subnet_id': subnet['id']}
        return (self._core_plugin.create_port(context, {
            'port':
            {'tenant_id': subnet['tenant_id'],
             'network_id': subnet['network_id'],
             'fixed_ips': [fixed_ip],
             'mac_address': utils.get_mac(self, context),
             'admin_state_up': True,
             'device_id': router.id,
             'device_owner': owner,
             'name': ''}}), [subnet], True)

    def _allocate_floatingip(self, context, obj):
        """
        1. mapping floatingip to the one of a pair of internal ips based on
           the vip function.
        2. add another ip of the ip pair to the secondaryip list of
           the external interface.

        obj example:
        {
            'floating_network_id': u'1c1dbecc-9dac-4311-a346-f147a04c8dc8',
            'router_id': None,
            'fixed_ip_address': None,
            'floating_ip_address': u'10.160.37.113',
            'tenant_id': u'3998b33381fb48f694369689065a3760',
            'status': 'DOWN',
            'port_id': None,
            'id': '5ec1b08b-77c1-4e39-80ac-224ee937ee9f'
        }

        The floatingip is a instance of neutron.db.l3_db.FloatingIP, example:
        {
            tenant_id=u'3998b33381fb48f694369689065a3760',
            id=u'25e1588a-5ec5-4fbc-bdef-eff8713da8f8',
            floating_ip_address=u'10.160.37.111',
            floating_network_id=u'1c1dbecc-9dac-4311-a346-f147a04c8dc8',
            floating_port_id=u'4b4120d4-77f9-4f82-b823-05876929a1c4',
            fixed_port_id=None,
            fixed_ip_address=None,
            router_id=None,
            last_known_router_id=None,
            status=u'DOWN'
        }
        """
        with context.session.begin(subtransactions=True):
            try:
                db_namespace = utils.add_vdom(self, context,
                                              tenant_id=obj['tenant_id'])

                db_fip = utils.add_record(self, context,
                                fortinet_db.Fortinet_FloatingIP_Allocation,
                                vdom=db_namespace.vdom,
                                floating_ip_address=obj['floating_ip_address'],
                                vip_name=obj['floating_ip_address'])
                mappedip = utils.get_ipaddr(db_fip.ip_subnet, 0)
                utils.add_vip(self, context,
                              vdom=const.EXT_VDOM,
                              name=db_fip.vip_name,
                              extip=db_fip.floating_ip_address,
                              extintf='any',
                              mappedip=mappedip)

                int_intf, ext_intf = utils.get_vlink_intf(self, context,
                                                       vdom=db_namespace.vdom)

                utils.add_fwpolicy(self, context,
                                   vdom=const.EXT_VDOM,
                                   dstintf=ext_intf,
                                   dstaddr=db_fip.vip_name,
                                   nat='enable')

                utils.add_routerstatic(self, context,
                                       vdom=const.EXT_VDOM,
                                       dst="%s 255.255.255.255" % mappedip,
                                       device=ext_intf,
                                       gateway=const.DEF_GW)

                utils.add_fwippool(self, context,
                                   name=db_fip.floating_ip_address,
                                   vdom=const.EXT_VDOM,
                                   startip=db_fip.floating_ip_address)

                utils.add_fwaddress(self, context,
                                    name=mappedip,
                                    vdom=const.EXT_VDOM,
                                    subnet="%s 255.255.255.255" % mappedip)

                db_fwpolicy = utils.add_fwpolicy(self, context,
                                   vdom=const.EXT_VDOM,
                                   srcintf=ext_intf,
                                   srcaddr=mappedip,
                                   dstintf=self._fortigate['ext_interface'],
                                   poolname=db_fip.floating_ip_address)
                utils.head_firewall_policy(self, context,
                                           vdom=const.EXT_VDOM,
                                           id=db_fwpolicy.edit_id)

                utils.add_fwippool(self, context,
                                   name=mappedip,
                                   vdom=db_namespace.vdom,
                                   startip=mappedip)
            except Exception as e:
                with excutils.save_and_reraise_exception():
                    utils._rollback_on_err(self, context, e)
        utils.update_status(self, context, t_consts.TaskStatus.COMPLETED)

    def _release_floatingip(self, context, id):
        """
        :param context:
        :param id: the floatingip id in neutron.db.l3_db.FloatingIP.
        {
                tenant_id=u'3998b33381fb48f694369689065a3760',
                id=u'25e1588a-5ec5-4fbc-bdef-eff8713da8f8',
                floating_ip_address=u'10.160.37.111',
                floating_network_id=u'1c1dbecc-9dac-4311-a346-f147a04c8dc8',
                floating_port_id=u'4b4120d4-77f9-4f82-b823-05876929a1c4',
                fixed_port_id=None,
                fixed_ip_address=None,
                router_id=None,
                last_known_router_id=None,
                status=u'DOWN'
        }
        :return:
        """
        with context.session.begin(subtransactions=True):
            l3db_fip = self._get_floatingip(context, id)
            tenant_id = l3db_fip.tenant_id
            db_namespace = fortinet_db.query_record(context,
                                    fortinet_db.Fortinet_ML2_Namespace,
                                    tenant_id=tenant_id)

            db_fip = fortinet_db.query_record(context,
                            fortinet_db.Fortinet_FloatingIP_Allocation,
                            floating_ip_address=l3db_fip.floating_ip_address,
                            allocated=True)
            if not db_fip or not db_namespace:
                return

            int_intf, ext_intf = utils.get_vlink_intf(self, context,
                                                     vdom=db_namespace.vdom)
            mappedip = utils.get_ipaddr(db_fip.ip_subnet, 0)

            utils.delete_fwippool(self, context,
                                  name=mappedip,
                                  vdom=db_namespace.vdom,
                                  startip=mappedip)

            utils.delete_fwpolicy(self, context,
                                  vdom=const.EXT_VDOM,
                                  srcintf=ext_intf,
                                  srcaddr=mappedip,
                                  dstintf=self._fortigate['ext_interface'],
                                  poolname=db_fip.floating_ip_address)

            utils.delete_fwaddress(self, context,
                                   name=mappedip,
                                   vdom=const.EXT_VDOM,
                                   subnet="%s 255.255.255.255" % mappedip)

            utils.delete_fwippool(self, context,
                                  name=db_fip.floating_ip_address,
                                  vdom=const.EXT_VDOM,
                                  startip=db_fip.floating_ip_address)

            utils.delete_routerstatic(self, context,
                                      vdom=const.EXT_VDOM,
                                      dst="%s 255.255.255.255" % mappedip,
                                      device=ext_intf,
                                      gateway=const.DEF_GW)

            utils.delete_fwpolicy(self, context,
                                  vdom=const.EXT_VDOM,
                                  dstintf=ext_intf,
                                  dstaddr=l3db_fip.floating_ip_address)

            utils.delete_vip(self, context,
                             vdom=const.EXT_VDOM,
                             name=db_fip.vip_name,
                             extip=db_fip.floating_ip_address,
                             extintf='any',
                             mappedip=mappedip)

            fortinet_db.delete_record(context,
                            fortinet_db.Fortinet_FloatingIP_Allocation,
                            vdom=db_namespace.vdom,
                            floating_ip_address=db_fip.floating_ip_address,
                            vip_name=db_fip.floating_ip_address)
            super(FortinetL3ServicePlugin, self).delete_floatingip(context, id)
            utils.delete_vlink(self, context, tenant_id)
            utils.delete_vdom(self, context, tenant_id=tenant_id)
